import tornado.web
from .base import HTTPHandler


class HTMLOpenHandler(HTTPHandler):
    def initialize(self, template=None, template_kwargs=None, **kwargs):
        super(HTMLOpenHandler, self).initialize()
        self.template = template
        self.template_kwargs = template_kwargs or {}

    def get(self, *args):
        '''Get the login page'''
        if not self.template:
            self.redirect('/')
        else:
            template = self.render_template(self.template, **self.template_kwargs)
            self.write(template)


class HTMLHandler(HTMLOpenHandler):
    @tornado.web.authenticated
    def get(self, *args):
        '''Get the login page'''
        if not self.template:
            self.redirect('/')
        else:
            template = self.render_template(self.template, **self.template_kwargs)
            self.write(template)



from .common import generateTornadoApplication
from ....ui.handlers.html import HTMLHandler, HTMLOpenHandler
from mock import MagicMock, patch
from tornado.web import HTTPError


class TestHTML:
    def setup(self):
        self.app, self.login_code = generateTornadoApplication()

    def test_HTMLOpenHandler(self):
        req = MagicMock()
        req.body = ''
        x = HTMLOpenHandler(self.app, req, template='404.html')
        x._transforms = []
        x.get_current_user = lambda: False

        x.template = '404.html'

        with patch.object(x, 'get_secure_cookie') as mget:
            mget.return_value = self.login_code
            x.get()

        x = HTMLOpenHandler(self.app, req, template=None)
        x._transforms = []
        x.get_current_user = lambda: False
        with patch.object(x, 'get_secure_cookie') as mget:
            mget.return_value = self.login_code
            x.get()

    def test_HTMLHandler(self):
        req = MagicMock()
        req.body = ''
        x = HTMLHandler(self.app, req, template='404.html')
        x._transforms = []
        x.get_current_user = lambda: False

        x.template = '404.html'
        try:
            with patch.object(x, 'get_secure_cookie') as mget:
                mget.return_value = self.login_code
                x.get()
            assert False
        except HTTPError:
            pass

        x = HTMLHandler(self.app, req, template=None)
        x._transforms = []
        x.get_current_user = lambda: True
        with patch.object(x, 'get_secure_cookie') as mget:
            mget.return_value = self.login_code
            x.get()

        x = HTMLHandler(self.app, req, template='404.html')
        x._transforms = []
        x.get_current_user = lambda: b'test'
        with patch.object(x, 'get_secure_cookie') as mget:
            mget.return_value = self.login_code
            x.get()



# for coverage
from ..config import *
from ..enums import TradingType


class TestConfig:
    def test_ExchangeConfig(self):
        x = ExchangeConfig()
        assert(x.exchange_types == [])
        assert(x.trading_type == TradingType.NONE)

    def test_BacktestConfig(self):
        x = BacktestConfig()
        assert(x)

    def test_RiskConfig(self):
        x = RiskConfig()
        assert(x.max_drawdown == 100.0)
        assert(x.max_risk == 100.0)
        assert(x.total_funds == 0.0)
        assert(x.trading_type == TradingType.NONE)

    def test_ExecutionConfig(self):
        x = ExecutionConfig()
        assert(x.trading_type == TradingType.NONE)

    def test_TradingEngineConfig(self):
        x = TradingEngineConfig()
        assert(x.type == TradingType.NONE)
        assert(x.print is False)



import time
import ujson
import base64
import hmac
import hashlib
import uuid
import secrets
import string
import os
import tornado.web
from ....trading import TradingEngine
from ....parser import parse_command_line_config


def generateApplication():
    config = parse_command_line_config(['--config=./config/sythetic.cfg'])

    # Instantiate trading engine
    #
    # The engine is responsible for managing the different components,
    # including the strategies, the bank/risk engine, and the
    # exchange/backtest engine.

    te = TradingEngine(config)

    # Run the live trading engine
    te.run()


def generateTornadoApplication():
    nonce = int(time.time() * 1000)
    encoded_payload = ujson.dumps({"nonce": nonce}).encode()
    b64 = base64.b64encode(encoded_payload)
    cookie_secret = hmac.new(str(uuid.uuid1()).encode(), b64, hashlib.sha384).hexdigest()
    login_code = ''.join(secrets.choice(string.ascii_letters + string.digits) for i in range(20))

    settings = {
        "debug": "True",
        "template_path": os.path.join(os.path.dirname(__file__), '../', '../', '../', 'ui', 'assets', 'templates'),
        "cookie_secret": cookie_secret,
        "login_code": login_code
        }

    app = tornado.web.Application(**settings)
    app._transforms = []
    return app, login_code



import tornado.gen
from tornado.concurrent import run_on_executor
from perspective import PerspectiveHTTPMixin
from .base import HTTPHandler
from aat.enums import ExchangeType_to_string


class ExchangesHandler(PerspectiveHTTPMixin, HTTPHandler):
    '''Server Handler
    Extends:
        HTTPHandler
    '''

    def initialize(self, trading_engine, psp_kwargs=None):
        self.te = trading_engine
        self.psp_kwargs = psp_kwargs or {}

    @run_on_executor
    def get_data(self, **psp_kwargs):
        exchanges = self.te.query.query_exchanges()
        msgs = [{'id': j + i * len(exchanges),
                 'name': ExchangeType_to_string(x['exchange']),
                 'instrument': y.to_dict(True, True)['underlying']}
                for i, x in enumerate(exchanges)
                for j, y in enumerate(x['instruments'])
                ]
        super(ExchangesHandler, self).loadData(data=msgs, **psp_kwargs)
        return super(ExchangesHandler, self).getData()

    @tornado.web.authenticated
    @tornado.gen.coroutine
    def get(self):
        dat = yield self.get_data(**self.psp_kwargs)
        self.write(dat)



import os
import os.path
from configparser import ConfigParser
from pydoc import locate
from .config import TradingEngineConfig, BacktestConfig, StrategyConfig, SyntheticExchangeConfig
from .enums import TradingType, InstrumentType, ExchangeType, PairType
from .exceptions import ConfigException
from .structs import Instrument
from .utils import str_to_exchange, set_verbose
from .logging import log


def parse_file_config(filename: str) -> TradingEngineConfig:
    if not os.path.exists(filename):
        raise ConfigException(f'File does not exist {filename}')
    config = TradingEngineConfig()
    c = ConfigParser()
    c.read(filename)

    general = c['general']
    exchange = c['exchange']
    strategy = c['strategy']
    risk = c['risk']
    default = c['DEFAULT']

    _parse_general(general, config)
    _parse_exchange(exchange, config)
    _parse_strategy(strategy, config)
    _parse_risk(risk, config)
    _parse_default(default, config)
    return config


def _parse_general(general, config) -> None:
    if 'TradingType' in general:
        if general['TradingType'].lower() == 'live':
            set_all_trading_types(TradingType.LIVE, config)
        elif general['TradingType'].lower() == 'simulation':
            set_all_trading_types(TradingType.SIMULATION, config)
        elif general['TradingType'].lower() == 'sandbox':
            set_all_trading_types(TradingType.SANDBOX, config)
        elif general['TradingType'].lower() == 'backtest':
            set_all_trading_types(TradingType.BACKTEST, config)
        else:
            raise ConfigException('Trading type not recognized : %s',
                                  general['TradingType'])
    else:
        raise ConfigException('TradingType unspecified')

    if 'verbose' in general:
        if int(general['verbose']) >= 1:
            set_verbose(int(general['verbose']))

    if 'print' in general:
        if general['print'] == '1':
            config.print = True


def _parse_exchange(exchange, config) -> None:
    if config.type == TradingType.LIVE:
        _parse_live_options(exchange, config)

    elif config.type == TradingType.SIMULATION:
        _parse_simulation_options(exchange, config)

    elif config.type == TradingType.SANDBOX:
        _parse_sandbox_options(exchange, config)

    elif config.type == TradingType.BACKTEST:
        _parse_backtest_options(exchange, config)

    else:
        ConfigException('No Trading Type specified in config!')


def _parse_strategy(strategy, config) -> None:
    strat_configs = []
    if 'strategies' not in strategy:
        raise ConfigException('No Strategies specified')

    for strat in strategy['strategies'].split('\n'):
        if strat == '':
            continue
        splits = [x for x in strat.split(',') if x]
        cls = locate(splits[0])

        args = []
        for x in splits[1:]:
            args.append(float(x.strip()) if x.strip().isdigit() else x)
        args = tuple(args)

        strat_configs.append(StrategyConfig(clazz=cls, args=args))
    config.strategy_options = strat_configs


def _parse_risk(risk, config) -> None:
    config.risk_options.max_drawdown = float(risk.get('max_drawdown', config.risk_options.max_drawdown))
    config.risk_options.max_risk = float(risk.get('max_risk', config.risk_options.max_risk))
    config.risk_options.total_funds = 0.0


def _parse_default(default, config) -> None:
    pass


def _parse_args_to_dict(argv: list) -> dict:
    ret = {}
    for item in argv[1:]:
        if '--' not in item:
            log.critical('Argument not recognized: %s', item)
        value = item.split('--')[1]
        if '=' in value:
            # = args
            splits = value.split('=')
            ret[splits[0]] = splits[-1]
        else:
            # single args
            if value.upper() in TradingType.members():
                ret['ttype'] = value
            elif value.upper() == 'VERBOSE':
                ret['verbose'] = True
            elif value.upper() == 'PRINT':
                ret['print'] = True
            else:
                log.critical('Argument not recognized: %s', item)
    return ret


def _parse_currencies(currencies):
    splits = [x.strip().upper().replace('-', '') for x in currencies.split(',')]
    ret = []
    for s in splits:
        ret.append(PairType.from_string(s))
    return ret


def _parse_exchanges(argv):
    return [str_to_exchange(x) for x in argv['exchanges'].split() if x]


def _parse_synthetic_config(config, argv):
    new_config = SyntheticExchangeConfig()

    new_config.exchange_types = config.exchange_options.exchange_types
    new_config.trading_type = config.exchange_options.trading_type
    new_config.currency_pairs = config.exchange_options.currency_pairs
    new_config.instruments = config.exchange_options.instruments

    if argv.get('direction'):
        new_config.direction = argv.get('direction')
    if argv.get('volatility'):
        new_config.volatility = argv.get('volatility')
    return new_config


def _parse_options(argv, config: TradingEngineConfig) -> None:
    if argv.get('exchanges'):
        config.exchange_options.exchange_types = _parse_exchanges(argv)
        for exchange in config.exchange_options.exchange_types:
            if config.type == TradingType.LIVE and exchange == ExchangeType.SYNTHETIC:
                raise ConfigException('Cannot run synthetic exchange in live mode!')
            elif exchange == ExchangeType.SYNTHETIC:
                config.exchange_options = _parse_synthetic_config(config, argv)
                config.exchange_options.exchange_type = ExchangeType.COINBASE
    else:
        raise ConfigException('No exchange set!')

    if argv.get('currency_pairs'):
        config.exchange_options.currency_pairs = _parse_currencies(argv.get('currency_pairs'))
        config.exchange_options.instruments = \
            [Instrument(type=InstrumentType.PAIR, underlying=p) for p in config.exchange_options.currency_pairs]


def _parse_live_options(argv, config: TradingEngineConfig) -> None:
    log.critical("\n\nWARNING: Live trading. money will be lost ;^)\n\n")
    _parse_options(argv, config)


def _parse_simulation_options(argv, config) -> None:
    log.critical("")
    log.critical("Simulation trading")
    log.critical("")
    _parse_options(argv, config)


def _parse_sandbox_options(argv, config) -> None:
    log.critical("")
    log.critical("Sandbox trading")
    log.critical("")
    _parse_options(argv, config)


def _parse_backtest_options(argv, config) -> None:
    log.critical("")
    log.critical("Backtesting")
    log.critical("")
    config.backtest_options = BacktestConfig()

    if argv.get('exchanges'):
        config.exchange_options.exchange_types = _parse_exchanges(argv)
        for exchange in config.exchange_options.exchange_types:
            if exchange == ExchangeType.SYNTHETIC:
                config.exchange_options = _parse_synthetic_config(config, argv)
                config.exchange_options.exchange_type = ExchangeType.COINBASE
    else:
        raise ConfigException('No exchange set!')

    if argv.get('currency_pairs'):
        config.exchange_options.currency_pairs = _parse_currencies(argv.get('currency_pairs'))


def parse_command_line_config(argv: list) -> TradingEngineConfig:
    # Every engine run requires a static config object
    argv = _parse_args_to_dict(argv)

    if argv.get('config'):
        config = parse_file_config(argv.get('config'))
    else:
        config = TradingEngineConfig()

        if 'live' == argv.get('ttype'):
            # WARNING: Live trading. money will be lost ;^)
            set_all_trading_types(TradingType.LIVE, config)
            _parse_live_options(argv, config)

        elif 'simulation' == argv.get('ttype'):
            # Trade against sandbox
            _parse_simulation_options(argv, config)
            set_all_trading_types(TradingType.SIMULATION, config)

        elif 'sandbox' == argv.get('ttype'):
            # Trade against sandbox
            _parse_sandbox_options(argv, config)
            set_all_trading_types(TradingType.SANDBOX, config)

        elif 'backtest' == argv.get('ttype'):
            # Backtest against trade data
            set_all_trading_types(TradingType.BACKTEST, config)
            _parse_backtest_options(argv, config)
        else:
            raise ConfigException('Trading Type not specified')

        if argv.get('verbose'):
            set_verbose()

        if argv.get('print'):
            config.print = True

    log.debug("Config : %s", str(config))

    return config


def set_all_trading_types(trading_type: TradingType,
                          config: TradingEngineConfig) -> None:
    '''Set all trading types to match'''
    config.type = trading_type
    config.exchange_options.trading_type = trading_type
    config.risk_options.trading_type = trading_type
    config.execution_options.trading_type = trading_type



import os
import os.path
import logging.config
import time

moment = time.strftime("%Y%m%d_%H%M%S", time.localtime())

if not os.path.isdir('./logs'):
    os.mkdir('./logs')

if not os.path.isdir('./logs' + moment):
    os.mkdir('./logs/' + moment)


LOGGING_CONFIG = {
    'version': 1,  # required
    'disable_existing_loggers': False,
    'formatters': {
        'simple': {
            'format': '%(asctime)s %(levelname)s -- %(message)s'
        },
        'whenAndWhere': {
            'format': '%(asctime)s\t%(levelname)s -- %(processName)s %(filename)s:%(lineno)s -- %(message)s'
        }
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'whenAndWhere'
        },
        'file': {
            'level': 'DEBUG',
            'class': 'logging.FileHandler',
            'formatter': 'whenAndWhere',
            'filename': 'logs/' + moment + '/out.log',
            'mode': 'w',
            'encoding': 'utf-8'
        },
    },
    'loggers': {
        '': {  # 'root' logger
            'level': 'CRITICAL',
            'handlers': ['console', 'file']
        },
    }
}

logging.config.dictConfig(LOGGING_CONFIG)
log = logging.getLogger('')  # factory method



from abc import ABCMeta, abstractmethod
from .structs import MarketData, TradeResponse
from .logging import log


class Callback(metaclass=ABCMeta):
    '''callback interface'''
    @abstractmethod
    def onTrade(self, data: MarketData):
        '''onTrade'''

    @abstractmethod
    def onOpen(self, data: MarketData):
        '''onOpen'''

    @abstractmethod
    def onFill(self, resp: TradeResponse):
        '''onFill'''

    @abstractmethod
    def onCancel(self, data: MarketData):
        '''onCancel'''

    @abstractmethod
    def onChange(self, data: MarketData):
        '''onChange'''

    @abstractmethod
    def onError(self, data: MarketData):
        '''onError'''

    def onStart(self):
        '''onStart'''
        pass

    def onExit(self):
        '''onExit'''
        pass

    def onAnalyze(self, engine):
        '''onAnalyze'''
        pass

    def onHalt(self, data):
        '''onHalt'''
        pass

    def onContinue(self, data):
        '''onContinue'''
        pass

    def callback(self):
        return self


class NullCallback(Callback):
    def __init__(self):
        pass

    def onTrade(self, data: MarketData) -> None:
        pass

    def onOpen(self, data: MarketData) -> None:
        pass

    def onFill(self, resp: TradeResponse) -> None:
        pass

    def onCancel(self, data: MarketData) -> None:
        pass

    def onChange(self, data: MarketData) -> None:
        pass

    def onError(self, data: MarketData) -> None:
        pass


class Print(Callback):
    def __init__(self,
                 onTrade=True,
                 onReceived=True,
                 onOpen=True,
                 onFill=True,
                 onCancel=True,
                 onChange=True,
                 onError=True):
        if not onTrade:
            setattr(self, 'onTrade', False)
        if not onReceived:
            setattr(self, 'onReceived', False)
        if not onOpen:
            setattr(self, 'onOpen', False)
        if not onFill:
            setattr(self, 'onFill', False)
        if not onCancel:
            setattr(self, 'onCancelled', False)
        if not onChange:
            setattr(self, 'onChange', False)
        if not onError:
            setattr(self, 'onError', False)

    def onTrade(self, data: MarketData) -> None:
        log.info(str(data))

    def onOpen(self, data: MarketData) -> None:
        log.info(str(data))

    def onFill(self, resp: TradeResponse) -> None:
        log.info(str(resp))

    def onCancel(self, data: MarketData) -> None:
        log.info(str(data))

    def onChange(self, data: MarketData) -> None:
        log.info(str(data))

    def onError(self, data: MarketData) -> None:
        log.info(str(data))

    def onAnalyze(self, data) -> None:
        log.info('Analysis')
        pass

    def onHalt(self, data) -> None:
        log.info('Halt')

    def onContinue(self, data) -> None:
        log.info('Continue')



import aiohttp
import json
from datetime import datetime
from functools import lru_cache
from typing import List
from .config import ExchangeConfig
from .enums import PairType, CurrencyType, ExchangeType, ExchangeType_to_string, TickType
from .market_data import MarketData
from .order_entry import OrderEntry
from .structs import Account, Instrument
from .exceptions import AATException
from .utils import findpath


class Exchange(MarketData, OrderEntry):
    def __init__(self,
                 exchange_type: ExchangeType,
                 options: ExchangeConfig,
                 query_engine=None) -> None:
        super(Exchange, self).__init__()
        self._options = options
        self._exchange = exchange_type
        self._query_engine = query_engine

    @lru_cache(None)
    def accounts(self):
        client = self.oe_client()
        if not client:
            return {}
        balances = client.fetch_balance()

        accounts = {}

        for jsn in balances['info']:
            if isinstance(jsn, str):
                currency = jsn
                jsn = balances['info'][jsn]
            else:
                currency = jsn['currency']

            currency = CurrencyType(currency)
            if 'balance' in jsn:
                balance = float(jsn['balance'])
            elif 'amount' in jsn:
                balance = float(jsn['amount'])
            elif 'available' in jsn:
                balance = jsn['available']
            else:
                raise AATException('Cant read!')

            id = jsn.get('id', jsn['currency'])
            account = Account(id=id,
                              currency=currency,
                              balance=balance,
                              exchange=self.exchange(),
                              value=-1,
                              asOf=datetime.now())
            accounts[currency] = account
        return accounts

    @lru_cache(None)
    def options(self) -> ExchangeConfig:
        return self._options

    @lru_cache(None)
    def exchange(self) -> ExchangeType:
        return self._exchange

    @lru_cache(None)
    def currencies(self) -> List[CurrencyType]:
        return [CurrencyType(x) for x in self.oe_client().fetch_curencies()]

    @lru_cache(None)
    def markets(self) -> List[Instrument]:
        # TODO derivatives
        return [Instrument(underlying=PairType.from_string(m['symbol'])) for m in self.oe_client().fetch_markets()]

    def ticker(self,
               instrument: Instrument = None,
               currency: CurrencyType = None):
        if instrument:
            return self.oe_client().fetchTicker(str(instrument.underlying))
        elif currency:
            inst = Instrument(underlying=PairType.from_string(currency.value + '/USD'))

            if inst in self.markets():
                return self.oe_client().fetchTicker(str(inst.underlying))
            else:
                for stable in ('USD', 'USDC', 'USDT'):
                    try:
                        inst = Instrument(underlying=PairType.from_string(currency.value + '/' + stable))
                        inst1, inst2, i1_inverted, i2_inverted = findpath(inst, self.markets())
                        broken = False
                    except (AATException, ValueError):
                        broken = True
                if broken:
                    return {'last': 0.0}
                inst1_t = self.oe_client().fetchTicker(str(inst1.underlying))
                inst2_t = self.oe_client().fetchTicker(str(inst2.underlying))
                if i1_inverted:
                    inst1_t['last'] = 1.0 / inst1_t['last']
                if i2_inverted:
                    inst2_t['last'] = 1.0 / inst2_t['last']
                px = inst1_t['last'] * inst2_t['last']
                ret = inst1_t
                for key in ret:
                    if key == 'info':
                        ret[key] = {}
                    elif key == 'last':
                        ret[key] = px
                    else:
                        ret[key] = None
                return ret

    def historical(self, timeframe='1m', since=None, limit=None):
        '''get historical data (for backtesting)'''
        import pandas as pd
        client = self.oe_client()
        dfs = [{'pair': str(symbol), 'exchange': ExchangeType_to_string(self.exchange()), 'data': client.fetch_ohlcv(symbol=str(symbol), timeframe=timeframe, since=since, limit=limit)}
               for symbol in self.options().currency_pairs]
        df = pd.io.json.json_normalize(dfs, 'data', ['pair', 'exchange'])
        df.columns = ['timestamp', 'open', 'high', 'low', 'close', 'volume', 'pair', 'exchange']
        df['timestamp'] = pd.to_datetime(df['timestamp'], unit='ms')
        df.set_index(['timestamp', 'pair'], inplace=True)
        df.sort_index(inplace=True)
        return df

    def orderBook(self, level=1):
        '''get order book'''
        return self.oe_client().getProductOrderBook(level=level)

    async def receive(self) -> None:
        async for msg in self.ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                self.callback_data(json.loads(msg.data))
            elif msg.type == aiohttp.WSMsgType.ERROR:
                self.callback(TickType.ERROR, msg.data)

    def callback_data(self, data) -> None:
        res = self.tickToData(data)
        if res is None:
            return

        if self._seqnum_enabled and res.type != TickType.HEARTBEAT:
            self.seqnum(res.sequence)

        if not self._running:
            pass

        if res.type != TickType.HEARTBEAT:
            self.callback(res.type, res)



import asyncio
import tornado
import uvloop
from .backtest import Backtest
from .callback import Print
from .config import TradingEngineConfig
from .enums import TradingType, Side, CurrencyType, TradeResult
from .execution import Execution
from .query import QueryEngine
from .risk import Risk
from .strategy import TradingStrategy
from .structs import TradeRequest, TradeResponse
from .ui.server import ServerApplication
from .utils import ex_type_to_ex, iterate_accounts
from .logging import log


class TradingEngine(object):
    def __init__(self, options: TradingEngineConfig, ui_handlers: list = None, ui_settings: dict = None) -> None:
        # save UI class for later
        ui_handlers = ui_handlers or []
        ui_settings = ui_settings or {}
        self._ui_handlers = ui_handlers
        self._ui_settings = ui_settings

        # trading type
        self.trading_type = options.type

        # instantiate exchange instance
        self.exchanges = {o: ex_type_to_ex(o)(o, options.exchange_options) for o in options.exchange_options.exchange_types}

        # lookup by exchange or currency
        self.accounts = {}
        for ex in self.exchanges.values():
            self.accounts[ex._exchange] = {}
            for account in ex.accounts().values():
                self.accounts[ex._exchange][account.currency] = account
                if account.currency not in self.accounts:
                    self.accounts[account.currency] = {}
                self.accounts[account.currency][ex._exchange] = account

        # get account information and balances
        for account in iterate_accounts(self.accounts):
            if self.trading_type == TradingType.BACKTEST:
                log.info(f'Adjusting account balance to 1,000 {account}')
                account.balance = 1000

            if account.currency == CurrencyType.USD:
                log.info(f'Adjusting account balance to 100,000 {account}')
                # if holding USD, add value
                account.balance = 100000
                account.value = account.balance
                options.risk_options.total_funds += account.balance
            else:
                log.info(f'Adjusting account balance to 1,000 {account}')
                account.balance = 1000

                # calculate USD value
                ex = self.exchanges[account.exchange]
                spot = ex.ticker(currency=account.currency)['last']
                options.risk_options.total_funds += account.balance * spot
                account.value = account.balance * spot

        log.info(self.accounts)
        log.info("Running with %.2f USD" % options.risk_options.total_funds)

        # instantiate backtest engine
        self.backtest = Backtest(options.backtest_options) if self.trading_type == TradingType.BACKTEST else None

        # instantiate riskengine
        self.risk = Risk(options.risk_options, self.exchanges, self.accounts)

        # instantiate execution engine
        self.execution = Execution(options.execution_options, self.exchanges, self.accounts)

        # instantiate query engine
        self.query = QueryEngine(trading_type=self.trading_type,
                                 exchanges=self.exchanges,
                                 pairs=options.exchange_options.currency_pairs,
                                 accounts=self.accounts,
                                 instruments={name:
                                              list(set(options.exchange_options.instruments).intersection(
                                                  ex.markets())) for name, ex in self.exchanges.items()},
                                 risk=self.risk,
                                 execution=self.execution)

        # register query hooks
        if self.trading_type in (TradingType.LIVE, TradingType.SIMULATION, TradingType.SANDBOX):
            for exc in self.exchanges.values():

                # Track my trades and cancels for future callbacks
                exc.onTrade(self.query.onTrade)
                exc.onCancel(self.query.onCancel)

                if options.print:
                    exc.registerCallback(Print(onTrade=True, onReceived=True, onOpen=True, onFill=True, onCancel=True, onChange=True, onError=False))
        else:
            self.backtest.onTrade(self.query.onTrade)
            self.backtest.onCancel(self.query.onCancel)
            self.backtest.registerCallback(Print())

        # register strategies from config
        for x in options.strategy_options:
            log.critical('Registering strategy: %s', str(x.clazz))
            x.kwargs['query'] = self.query
            x.kwargs['exchanges'] = self.exchanges
            self.registerStrategy(x.clazz(*x.args, **x.kwargs))

        # actively trading or halted?
        self._trading = True  # type: bool

    def haltTrading(self):
        self._trading = False
        for strat in self.query.strategies:
            strat.onHalt()

    def continueTrading(self):
        self._trading = True
        for strat in self.query.strategies:
            strat.onContinue()

    def registerStrategy(self, strat: TradingStrategy):
        if self.trading_type in (TradingType.LIVE, TradingType.SIMULATION, TradingType.SANDBOX):
            # register for exchange data
            for ex in self.exchanges.values():
                ex.registerCallback(strat.callback())
        else:
            # register for backtest data
            self.backtest.registerCallback(strat.callback())

        # add to tickables
        self.query.registerStrategy(strat)

        # give self to strat so it can request trading actions
        strat.setEngine(self)

    def run(self):
        if self.trading_type in (TradingType.LIVE, TradingType.SIMULATION, TradingType.SANDBOX):
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

            loop = tornado.platform.asyncio.AsyncIOMainLoop().install()

            port = 8081
            self.application = ServerApplication(self,
                                                 extra_handlers=self._ui_handlers,
                                                 custom_settings=self._ui_settings)

            # trigger starts
            for strat in self.query.strategies:
                strat.onStart()

            # run on exchange
            async def _run():
                await asyncio.wait([ex.run(self) for ex in self.exchanges.values()])

            # get event loop
            loop = asyncio.get_event_loop()

            # hook in tornado to asyncio
            log.critical('')
            log.critical('Server listening on port: %s', port)
            log.critical('')
            self.application.listen(port)

            # run asyncio loop
            loop.create_task(_run())
            loop.run_forever()

        else:
            # trigger starts
            for strat in self.query.strategies:
                strat.onStart()

            # let backtester run
            self.backtest.run(self)

    def terminate(self):
        for strat in self._strats:
            strat.onExit()

    def _request(self,
                 side: Side,
                 req: TradeRequest,
                 strat=None):
        self.query.push_tradereq(req)

        if not self._trading:
            # not allowed to trade right now
            log.info('Trading not allowed')
            resp = TradeResponse(request=req,
                                 side=req.side,
                                 exchange=req.exchange,
                                 volume=0.0,
                                 price=0.0,
                                 instrument=req.instrument,
                                 strategy=strat,
                                 order_id='')
        else:
            # get risk report
            resp = self.risk.request(req)

            if resp.risk_check:
                log.info(f'Risk check passed: {resp}')
                # if risk passes, let execution execute
                # Note that the Risk+Execution must be atomic
                # so that the risk module can keep an up-to-date
                # calculation of outstanding risk
                resp = self.execution.request(resp)

                if resp.status == TradeResult.PENDING:
                    # listen for later fill/cancel
                    log.info(f'Order pending: {resp}')
                    self.query.newPending(resp)

                elif resp.status == TradeResult.REJECTED:
                    # send response
                    log.info(f'Trade rejected: {resp}')
                    resp = TradeResponse(request=resp,
                                         side=resp.side,
                                         exchange=req.exchange,
                                         volume=0.0,
                                         price=req.price,
                                         time=req.time,
                                         instrument=req.instrument,
                                         status=TradeResult.REJECTED,
                                         strategy=strat,
                                         order_id='')

                    # take the risk off the books
                    self.risk.cancel(resp)

                elif resp.status == TradeResult.FILLED:
                    if self.trading_type in (TradingType.SIMULATION, TradingType.BACKTEST):
                        # adjust response with slippage and transaction cost modeling
                        resp = strat.slippage(resp)

                        # adjust response with slippage and transaction cost modeling
                        resp = strat.transactionCost(resp)

                        # mark as pending
                        self.query.newPending(resp)

                        # force run through query engine
                        self.query.onFill(resp)

                    log.info(f'Trade filled: {resp}')
                    log.info("Slippage - %s" % resp.slippage)
                    log.info("TXN cost - %s" % resp.transaction_cost)

                    # let risk update according to execution details
                    self.risk.update(resp)

            else:
                log.info('Risk check failed')
                resp = TradeResponse(request=resp,
                                     exchange=req.exchange,
                                     side=resp.side,
                                     volume=0.0,
                                     price=0.0,
                                     time=req.time,
                                     instrument=req.instrument,
                                     status=TradeResult.REJECTED,
                                     strategy=strat,
                                     order_id='')

        self.query.push_traderesp(resp)
        self.query.update_positions(resp)
        return resp

    def request(self, req: TradeRequest, strat=None):
        req.strategy = strat
        return self._request(side=req.side,
                             req=req,
                             strat=strat)

    def cancel(self, resp: TradeResponse, strat=None):
        resp = self.execution.cancel(resp)
        return resp

    def cancelAll(self, strat=None):
        resp = self.execution.cancelAll(self._pending)
        return resp



class TestDefine:
    def setup(self):
        pass
        # setup() before each test method

    def teardown(self):
        pass
        # teardown() after each test method

    @classmethod
    def setup_class(cls):
        pass
        # setup_class() before any methods in this class

    @classmethod
    def teardown_class(cls):
        pass
        # teardown_class() after any methods in this class

    def test_exchange_endpoint(self):
        from ..define import EXCHANGE_MARKET_DATA_ENDPOINT
        from ..enums import ExchangeType, TradingType

        assert EXCHANGE_MARKET_DATA_ENDPOINT(ExchangeType.COINBASE, TradingType.SANDBOX) \
            == 'wss://ws-feed-public.sandbox.pro.coinbase.com'
        assert EXCHANGE_MARKET_DATA_ENDPOINT(ExchangeType.COINBASE, TradingType.LIVE) \
            == 'wss://ws-feed.pro.coinbase.com'
        assert EXCHANGE_MARKET_DATA_ENDPOINT(ExchangeType.GEMINI, TradingType.SANDBOX) \
            == 'wss://api.sandbox.gemini.com/v1/marketdata/%s?heartbeat=true'
        assert EXCHANGE_MARKET_DATA_ENDPOINT(ExchangeType.GEMINI, TradingType.LIVE) \
            == 'wss://api.gemini.com/v1/marketdata/%s?heartbeat=true'
        assert EXCHANGE_MARKET_DATA_ENDPOINT(ExchangeType.KRAKEN, TradingType.SANDBOX) \
            == 'wss://ws-beta.kraken.com'
        assert EXCHANGE_MARKET_DATA_ENDPOINT(ExchangeType.KRAKEN, TradingType.LIVE) \
            == 'wss://ws.kraken.com'
        assert EXCHANGE_MARKET_DATA_ENDPOINT(ExchangeType.POLONIEX, TradingType.SANDBOX) \
            == ''
        assert EXCHANGE_MARKET_DATA_ENDPOINT(ExchangeType.POLONIEX, TradingType.LIVE) \
            == 'wss://api2.poloniex.com'



class TestExchange:
    def setup(self):
        pass
        # setup() before each test method

    def teardown(self):
        pass
        # teardown() after each test method

    @classmethod
    def setup_class(cls):
        pass
        # setup_class() before any methods in this class

    @classmethod
    def teardown_class(cls):
        pass
        # teardown_class() after any methods in this class

    def test_init(self):
        from ..config import ExchangeConfig
        from ..exchange import Exchange

        ec = ExchangeConfig()

        try:
            e = Exchange(ec)
            assert e
            assert False
        except:
            assert True



from ..backtest import *
from ..callback import *
from ..config import *
from ..data_source import *
from ..define import *
from ..enums import *
from ..exceptions import *
from ..exchange import *
from ..exchanges.coinbase import *
from ..exchanges.gemini import *
from ..exchanges.kraken import *
from ..exchanges.poloniex import *
from ..execution import *
from ..logging import *
from ..market_data import *
from ..order_book import *
from ..order_entry import *
from ..parser import *
from ..query import *
from ..risk import *
from ..strategy import *
from ..structs import *
from ..trading import *
from ..ui.server import *
from ..ui.handlers.accounts import *
from ..ui.handlers.base import *
from ..ui.handlers.html import *
from ..ui.handlers.instruments import *
from ..ui.handlers.trades import *
from ..utils import *



import aiohttp
import base64
import json
import hashlib
import hmac
import time
from aiostream import stream
from datetime import datetime
from functools import lru_cache
from ..define import EXCHANGE_MARKET_DATA_ENDPOINT
from ..enums import TickType, TickType_from_string, PairType
from ..exchange import Exchange
from ..logging import log
from ..structs import MarketData, Instrument
from ..utils import str_to_side


class GeminiExchange(Exchange):
    @lru_cache(None)
    def subscription(self):
        return [json.dumps({"type": "subscribe", "product_id": x.value[0].value + x.value[1].value}) for x in self.options().currency_pairs]

    @lru_cache(None)
    def heartbeat(self):
        return ''

    async def run(self, engine) -> None:
        options = self.options()
        # private events
        gemini_api_key = self.oe_client().apiKey
        gemini_api_secret = self.oe_client().secret.encode()

        payload = {"request": "/v1/order/events", "nonce": int(time.time() * 1000)}
        encoded_payload = json.dumps(payload).encode()
        b64 = base64.b64encode(encoded_payload)
        signature = hmac.new(gemini_api_secret, b64, hashlib.sha384).hexdigest()

        session = aiohttp.ClientSession(headers={
            'X-GEMINI-PAYLOAD': b64.decode(),
            'X-GEMINI-APIKEY': gemini_api_key,
            'X-GEMINI-SIGNATURE': signature
        })

        # startup and redundancy
        log.info('Starting....')
        self.ws = [await session.ws_connect(EXCHANGE_MARKET_DATA_ENDPOINT(self.exchange(), options.trading_type) % x) for x in self.subscription()]
        private_events = await session.ws_connect("wss://api.gemini.com/v1/order/events")
        self.ws.append(private_events)

        # set subscription for each ws
        for i, sub in enumerate(self.subscription()):
            self.ws[i]._subscription = sub
            log.info(f'Sending Subscription {sub}')

        log.info(f'Connected: {self.exchange()}')
        log.info('')
        log.critical(f'Starting algo trading: {self.exchange()}')
        try:
            while True:
                await self.receive()

        except KeyboardInterrupt:
            log.critical('Terminating program')
            engine.terminate()
            return

    async def receive(self) -> None:
        '''gemini has its own receive method because it uses 1 connection per symbol instead of multiplexing'''
        async def get_data_sub_pair(ws, sub=None):
            async for ret in ws:
                yield ret, sub

        # add one for private stream
        async for val in stream.merge(*[get_data_sub_pair(self.ws[i], sub) for i, sub in enumerate(self.subscription() + [None])]):
            jsn = json.loads(val[0].data)

            if isinstance(jsn, dict) and 'events' in jsn:
                events = jsn.get('events', [])
            elif not isinstance(jsn, list):
                events = [jsn]
            else:
                events = jsn

            if val[1]:
                # data stream
                pair = json.loads(val[1]).get('product_id')
            else:
                # private events
                pair = None

            for item in events:
                if item.get('type', 'subscription_ack') in ('subscription_ack', 'heartbeat'):
                    # can skip these
                    continue
                if item.get('type') == 'accepted':
                    # can ignore these as well, will have a fill and/or booked
                    # https://docs.gemini.com/websocket-api/#workflow
                    continue
                if item.get('type') == 'closed':
                    # can ignore these as well, will have a fill or cancelled
                    # https://docs.gemini.com/websocket-api/#workflow
                    continue

                if pair is None:
                    # private events
                    pair = item['symbol']

                item['symbol'] = pair
                res = self.tickToData(item)

                if not self._running:
                    pass

                if res.type != TickType.HEARTBEAT:
                    self.callback(res.type, res)

    def tickToData(self, jsn: dict) -> MarketData:
        order_id = jsn.get('order_id', '') or str(jsn.get('tid', ''))
        time = datetime.now()
        price = float(jsn.get('price', 'nan'))
        volume = float(jsn.get('amount', 0.0))

        s = jsn.get('type').upper()
        if s in ('BLOCK_TRADE', 'FILL'):  # Market data can't trigger fill event
            typ = TickType.TRADE
        elif s in ('AUCTION_INDICATIVE', 'AUCTION_OPEN', 'BOOKED', 'INITIAL'):
            typ = TickType.OPEN
        elif s in ('CANCELLED',):
            typ = TickType.CANCEL
        else:
            typ = TickType_from_string(s)
        delta = float(jsn.get('delta', 0.0))

        if typ == TickType.CHANGE and not volume:
            delta = float(jsn.get('delta', 'nan'))
            volume = delta
            # typ = self.reasonToTradeType(reason)

        side = str_to_side(jsn.get('side', ''))
        remaining_volume = float(jsn.get('remaining', jsn.get('remaining_amount', 'nan')))
        sequence = -1

        if 'symbol' not in jsn:
            return

        currency_pair = PairType.from_string(jsn.get('symbol'))
        instrument = Instrument(underlying=currency_pair)

        ret = MarketData(order_id=order_id,
                         time=time,
                         volume=volume,
                         price=price,
                         type=typ,
                         instrument=instrument,
                         remaining=remaining_volume,
                         side=side,
                         exchange=self.exchange(),
                         sequence=sequence)
        return ret



from abc import ABCMeta, abstractmethod
from .callback import Callback
from .structs import TradeRequest, TradeResponse
from .enums import TickType
from .exceptions import CallbackException


class DataSource(metaclass=ABCMeta):
    pass


class RestAPIDataSource(DataSource):
    @abstractmethod
    def accounts(self):
        '''get account information'''

    @abstractmethod
    def buy(self, req: TradeRequest) -> TradeResponse:
        '''execute a buy order'''

    @abstractmethod
    def sell(self, req: TradeRequest) -> TradeResponse:
        '''execute a sell order'''

    @abstractmethod
    def cancel(self, resp: TradeResponse) -> None:
        '''cancel an order'''

    @abstractmethod
    def cancelAll(self) -> None:
        '''cancel all orders'''

    @abstractmethod
    def orderBook(self):
        '''return the order book'''

    @abstractmethod
    def historical(self, timeframe='1m', since=None, limit=None):
        '''get historical data (for backtesting)'''


class StreamingDataSource(DataSource):
    def __init__(self, *args, **kwargs) -> None:
        self._running = False
        self._callbacks = {TickType.TRADE: [],
                           TickType.ERROR: [],
                           TickType.OPEN: [],
                           TickType.FILL: [],
                           TickType.CANCEL: [],
                           TickType.CHANGE: [],
                           TickType.ANALYZE: [],
                           TickType.HALT: [],
                           TickType.CONTINUE: []}

    @abstractmethod
    async def run(self, engine) -> None:
        '''run the exchange'''

    @abstractmethod
    async def close(self) -> None:
        '''close the websocket'''

    @abstractmethod
    def seqnum(self, number: int):
        '''manage sequence numbers'''

    @abstractmethod
    def receive(self):
        '''receive data and call callbacks'''

    def callback(self, field: str, data, *args, **kwargs) -> None:
        for cb in self._callbacks[field]:
            cb(data, *args, **kwargs)

    # Data functions
    @abstractmethod
    def tickToData(self, jsn):
        '''convert json to market data based on fields'''

    def onTrade(self, callback: Callback) -> None:
        self._callbacks[TickType.TRADE].append(callback)

    def onOpen(self, callback: Callback) -> None:
        self._callbacks[TickType.OPEN].append(callback)

    def onFill(self, callback: Callback) -> None:
        self._callbacks[TickType.FILL].append(callback)

    def onCancel(self, callback: Callback) -> None:
        self._callbacks[TickType.CANCEL].append(callback)

    def onChange(self, callback: Callback) -> None:
        self._callbacks[TickType.CHANGE].append(callback)

    def onError(self, callback: Callback) -> None:
        self._callbacks[TickType.ERROR].append(callback)

    def onExit(self, callback: Callback) -> None:
        self._callbacks[TickType.EXIT].append(callback)

    def onAnalyze(self, callback: Callback) -> None:
        self._callbacks[TickType.ANALYZE].append(callback)

    def onHalt(self, callback: Callback) -> None:
        self._callbacks[TickType.HALT].append(callback)

    def onContinue(self, callback: Callback) -> None:
        self._callbacks[TickType.CONTINUE].append(callback)

    def registerCallback(self, callback: Callback) -> None:
        if not isinstance(callback, Callback):
            raise CallbackException(f'{callback} is not an instance of class Callback')
        for att in ['onTrade',
                    'onOpen',
                    'onFill',
                    'onCancel',
                    'onChange',
                    'onError',
                    'onAnalyze',
                    'onHalt',
                    'onContinue']:
            if hasattr(callback, att):
                getattr(self, att)(getattr(callback, att))



import aiohttp
from abc import abstractmethod
from .data_source import StreamingDataSource
from .define import EXCHANGE_MARKET_DATA_ENDPOINT
from .structs import MarketData
from .logging import log


class MarketData(StreamingDataSource):
    def __init__(self, *args, **kwargs) -> None:
        super(MarketData, self).__init__()
        self._lastseqnum = -1
        self._missingseqnum = set()  # type: set
        self._seqnum_enabled = False

    @abstractmethod
    def subscription(self):
        '''subscription for websocket'''

    @abstractmethod
    def heartbeat(self):
        '''heartbeat for websocket'''

    def seqnum(self, number: int) -> None:
        if self._lastseqnum == -1:
            # first seen
            self._lastseqnum = number
            return

        if number != self._lastseqnum + 1 and number not in self._missingseqnum:
            log.error('ERROR: Missing sequence number/s: %s' % ','.join(
                str(x) for x in range(self._lastseqnum + 1, number + 1)))
            self._missingseqnum.update(
                x for x in range(self._lastseqnum + 1, number + 1))
            log.error(self._missingseqnum)

        if number in self._missingseqnum:
            self._missingseqnum.remove(number)
            log.warning('INFO: Got out of order data for seqnum: %s' % number)

        else:
            self._lastseqnum = number

    async def close(self) -> None:
        '''close the websocket'''
        await self.ws.close()

    async def run(self, engine) -> None:
        options = self.options()
        session = aiohttp.ClientSession()

        while True:
            # startup and redundancy
            log.info('Starting....')
            self.ws = await session.ws_connect(EXCHANGE_MARKET_DATA_ENDPOINT(self.exchange(), options.trading_type))
            log.info(f'Connected: {self.exchange()}')

            for sub in self.subscription():
                await self.ws.send_str(sub)
                log.info('Sending Subscription %s' % sub)

            if self.heartbeat():
                await self.ws.send_str(self.heartbeat())
                log.info('Sending Heartbeat %s' % self.heartbeat())

            log.info('')
            log.critical(f'Starting algo trading: {self.exchange()}')
            try:
                while True:
                    await self.receive()

            except KeyboardInterrupt:
                log.critical('Terminating program')
                return

    @abstractmethod
    def tickToData(self, jsn: dict) -> MarketData:
        pass



from datetime import datetime


class TestUtils:
    def test_parse_date(self):
        from datetime import datetime
        from ..utils import parse_date
        gold = datetime(2016, 11, 16, 0, 0)
        date1 = parse_date('1479272400.0')
        date2 = parse_date('2016-11-16T00:00:00.000000Z')
        print(gold)
        print(date1)
        print(date2)

        assert gold == date1 == date2

    def test_ex_type_to_ex(self):
        from ..utils import ex_type_to_ex
        from ..enums import ExchangeType
        from ..exchanges.coinbase import CoinbaseExchange
        assert ex_type_to_ex(ExchangeType.COINBASE) == CoinbaseExchange

    def test_set_verbose(self):
        import logging
        from ..utils import set_verbose
        from ..logging import log
        set_verbose(2)

        assert(log.level == logging.DEBUG)

    def test_get_keys_from_environment(self):
        from ..utils import get_keys_from_environment
        import os
        os.environ['TEST_API_KEY'] = 'test'
        os.environ['TEST_API_SECRET'] = 'test'
        os.environ['TEST_API_PASS'] = 'test'
        one, two, three = get_keys_from_environment('TEST')
        assert(one == 'test')
        assert(two == 'test')
        assert(three == 'test')

    def test_str_to_side(self):
        from ..utils import str_to_side
        from ..enums import Side
        assert(str_to_side('BUY') == Side.BUY)
        assert(str_to_side('SELL') == Side.SELL)
        assert(str_to_side('OTHER') == Side.NONE)

    def test_str_to_order_type(self):
        from ..utils import str_to_order_type
        from ..enums import OrderType
        assert(str_to_order_type('MARKET') == OrderType.MARKET)
        assert(str_to_order_type('LIMIT') == OrderType.LIMIT)
        assert(str_to_order_type('OTHER') == OrderType.NONE)

    def test_str_to_exchange(self):
        from ..utils import str_to_exchange
        from ..enums import ExchangeType
        assert(str_to_exchange('coinbase') == ExchangeType.COINBASE)
        assert(str_to_exchange('gemini') == ExchangeType.GEMINI)
        assert(str_to_exchange('kraken') == ExchangeType.KRAKEN)
        assert(str_to_exchange('poloniex') == ExchangeType.POLONIEX)

    def test_trade_req_to_params(self):
        from ..utils import trade_req_to_params
        from ..structs import TradeRequest, Instrument, ExchangeType
        from ..enums import Side, OrderType, PairType

        t = TradeRequest(side=Side.BUY,
                         volume=1.0,
                         price=1.0,
                         instrument=Instrument(underlying=PairType.BTCUSD),
                         exchange=ExchangeType.COINBASE,
                         order_type=OrderType.LIMIT,
                         time=datetime.now())

        ret = trade_req_to_params(t)

        assert ret['symbol'] == 'BTC/USD'
        assert ret['side'] == 'buy'
        assert ret['type'] == 'limit'
        assert ret['amount'] == 1.0
        assert ret['price'] == 1.0



from mock import patch, MagicMock


class TestExchange:
    def test_init(self):
        from ...config import ExchangeConfig
        from ...exchanges.coinbase import CoinbaseExchange
        from ...enums import ExchangeType

        ec = ExchangeConfig()
        ec.exchange_type = ExchangeType.COINBASE
        e = CoinbaseExchange(ExchangeType.COINBASE, ec)
        e._running = True
        assert e

    def test_receive(self):
        from ...config import ExchangeConfig
        from ...exchanges.coinbase import CoinbaseExchange
        from ...enums import TickType, ExchangeType

        with patch('os.environ'), patch('ccxt.coinbasepro'):
            ec = ExchangeConfig()
            ec.exchange_type = ExchangeType.COINBASE
            e = CoinbaseExchange(ExchangeType.COINBASE, ec)
            e._running = True
            assert e

            e.ws = MagicMock()

            with patch('json.loads') as m1:
                for i, val in enumerate([TickType.TRADE,
                                         TickType.OPEN,
                                         TickType.FILL,
                                         TickType.CANCEL,
                                         TickType.CHANGE,
                                         TickType.ERROR]):
                    m1.return_value = {'type': val,
                                       'sequence': i,
                                       'time': '2017-02-19T18:52:17.088000Z',
                                       'product_id': 'BTCUSD'}
                    e.receive()

    # def test_seqnum_fix(self):
    #     from ...lib.config import ExchangeConfig
    #     from ...lib.exchanges.gdax import GDAXExchange
    #     from ...lib.enums import TickType, ExchangeType

    #     with patch('os.environ'), patch('gdax.AuthenticatedClient'):
    #         ec = ExchangeConfig()
    #         ec.exchange_type = ExchangeType.GDAX
    #         e = GDAXExchange(ec)
    #         e._running = True
    #         assert e

    #         e.ws = MagicMock()

    #         with patch('json.loads') as m1:
    #             m1.return_value = {'type': TickType.TRADE,
    #                                'sequence': 0,
    #                                'time': '2017-02-19T18:52:17.088000Z',
    #                                'product_id': 'BTCUSD'}
    #             e.receive()
    #             for i, val in enumerate([TickType.TRADE,
    #                                      TickType.RECEIVED,
    #                                      TickType.OPEN,
    #                                      TickType.DONE,
    #                                      TickType.CHANGE,
    #                                      TickType.ERROR]):
    #                 m1.return_value = {'type': val,
    #                                    'sequence': 6-i,
    #                                    'time': '2017-02-19T18:52:17.088000Z',
    #                                    'product_id': 'BTCUSD'}
    #                 if i != 0:
    #                     assert e._missingseqnum
    #                 e.receive()
    #             assert e._missingseqnum == set()



import base64
import hashlib
import hmac
import os
import os.path
import logging
import secrets
import string
import time
import tornado.ioloop
import tornado.web
import ujson
import uuid
from .handlers.accounts import AccountsHandler
from .handlers.exchanges import ExchangesHandler
from .handlers.instruments import InstrumentsHandler
from .handlers.last_price import LastPriceHandler
from .handlers.login import LoginHandler, LogoutHandler
from .handlers.strategies import StrategiesHandler
from .handlers.strategy_trade_request import StrategyTradeRequestHandler
from .handlers.strategy_trade_response import StrategyTradeResponseHandler
from .handlers.trades import TradesHandler
from .handlers.html import HTMLHandler, HTMLOpenHandler
from ..logging import log


class ServerApplication(tornado.web.Application):
    def __init__(self,
                 trading_engine,
                 extra_handlers=None,
                 custom_settings=None,
                 debug=True,
                 cookie_secret=None,
                 *args,
                 **kwargs):
        root = os.path.join(os.path.dirname(__file__), 'assets')
        static = os.path.join(root, 'static')

        logging.getLogger('tornado.access').disabled = False

        if not cookie_secret:
            nonce = int(time.time() * 1000)
            encoded_payload = ujson.dumps({"nonce": nonce}).encode()
            b64 = base64.b64encode(encoded_payload)
            cookie_secret = hmac.new(str(uuid.uuid1()).encode(), b64, hashlib.sha384).hexdigest()

        login_code = ''.join(secrets.choice(string.ascii_letters + string.digits) for i in range(20))
        log.critical(f'\n**********\nLogin code: {login_code}\n**********')

        settings = {
            "cookie_secret": cookie_secret,
            "login_url": "/login",
            "login_code": login_code,
            "debug": debug,
            "template_path": os.path.join(root, 'templates'),
        }
        settings.update(custom_settings or {})
        extra_handlers = extra_handlers or []
        for route, handler, h_kwargs in extra_handlers:
            if 'trading_engine' in h_kwargs:
                h_kwargs['trading_engine'] = trading_engine

        super(ServerApplication, self).__init__(
            extra_handlers + [
                (r"/api/v1/json/accounts", AccountsHandler, {'trading_engine': trading_engine, 'psp_kwargs': {'index': 'id'}}),
                (r"/api/v1/arrow/accounts", AccountsHandler, {'trading_engine': trading_engine, 'psp_kwargs': {'index': 'id', 'transfer_as_arrow': True}}),
                (r"/api/v1/json/exchanges", ExchangesHandler, {'trading_engine': trading_engine, 'psp_kwargs': {'index': 'id'}}),
                (r"/api/v1/arrow/exchanges", ExchangesHandler, {'trading_engine': trading_engine, 'psp_kwargs': {'index': 'id', 'transfer_as_arrow': True}}),
                (r"/api/v1/json/instruments", InstrumentsHandler, {'trading_engine': trading_engine, 'psp_kwargs': {'index': 'underlying'}}),
                (r"/api/v1/arrow/instruments", InstrumentsHandler, {'trading_engine': trading_engine, 'psp_kwargs': {'index': 'underlying', 'transfer_as_arrow': True}}),
                (r"/api/v1/json/strategies", StrategiesHandler, {'trading_engine': trading_engine,
                                                                 'psp_kwargs': {'view': 'hypergrid'}}),
                (r"/api/v1/arrow/strategies", StrategiesHandler, {'trading_engine': trading_engine,
                                                                  'psp_kwargs': {'view': 'hypergrid', 'transfer_as_arrow': True}}),
                (r"/api/v1/json/strategy-trade-requests", StrategyTradeRequestHandler, {'trading_engine': trading_engine,
                                                                                        'psp_kwargs': {'index': 'time', 'view': 'hypergrid'}}),
                (r"/api/v1/arrow/strategy-trade-requests", StrategyTradeRequestHandler, {'trading_engine': trading_engine,
                                                                                         'psp_kwargs': {'index': 'time', 'view': 'hypergrid', 'transfer_as_arrow': True}}),
                (r"/api/v1/json/strategy-trade-responses", StrategyTradeResponseHandler, {'trading_engine': trading_engine,
                                                                                          'psp_kwargs': {'index': 'time', 'view': 'hypergrid'}}),
                (r"/api/v1/arrow/strategy-trade-responses", StrategyTradeResponseHandler, {'trading_engine': trading_engine,
                                                                                           'psp_kwargs': {'index': 'time', 'view': 'hypergrid', 'transfer_as_arrow': True}}),
                (r"/api/v1/json/last-price-all", LastPriceHandler, {'trading_engine': trading_engine,
                                                                    'psp_kwargs': {'index': 'instrument', 'view': 'hypergrid'}}),
                (r"/api/v1/arrow/last-price-all", LastPriceHandler, {'trading_engine': trading_engine,
                                                                     'psp_kwargs': {'index': 'instrument', 'view': 'hypergrid', 'transfer_as_arrow': True}}),
                (r"/api/v1/json/trades", TradesHandler, {'trading_engine': trading_engine,
                                                         'psp_kwargs': {'index': 'time', 'view': 'hypergrid', 'limit': 100}}),
                (r"/api/v1/arrow/trades", TradesHandler, {'trading_engine': trading_engine,
                                                          'psp_kwargs': {'index': 'time', 'view': 'hypergrid', 'limit': 100, 'transfer_as_arrow': True}}),
                (r"/static/(.*)", tornado.web.StaticFileHandler, {"path": static}),
                (r"/api/v1/login", LoginHandler, {}),
                (r"/api/v1/logout", LogoutHandler, {}),
                (r"/login", HTMLOpenHandler, {'template': '404.html'}),
                (r"/logout", HTMLHandler, {'template': '404.html'}),
                (r"/(.*)", HTMLOpenHandler, {'template': '404.html'})
            ], **settings)



from functools import lru_cache
from enum import Enum
from ._enums import (  # noqa: F401
    TickType,
    TickTypes,
    TickType_to_string,
    TickType_from_string,
    ExchangeType,
    ExchangeTypes,
    ExchangeType_to_string,
    ExchangeType_from_string,
    )


class BaseEnum(Enum):
    @classmethod
    def members(cls):
        return cls.__members__.keys()


class TradingType(BaseEnum):
    NONE = 'NONE'
    SANDBOX = 'SANDBOX'
    LIVE = 'LIVE'
    BACKTEST = 'BACKTEST'
    SIMULATION = 'SIMULATION'


class CurrencyType(BaseEnum):
    NONE = 'NONE'  # special, dont use

    USD = 'USD'
    EUR = 'EUR'
    GBP = 'GBP'

    USDC = 'USDC'
    USDT = 'USDT'

    BAT = 'BAT'
    BCH = 'BCH'
    BTC = 'BTC'
    CVC = 'CVC'
    DAI = 'DAI'
    DNT = 'DNT'
    EOS = 'EOS'
    ETC = 'ETC'
    ETH = 'ETH'
    GNT = 'GNT'
    LINK = 'LINK'
    LOOM = 'LOOM'
    LTC = 'LTC'
    MANA = 'MANA'
    REP = 'REP'
    XLM = 'XLM'
    XRP = 'XRP'
    ZEC = 'ZEC'
    ZRX = 'ZRX'


def _joiner(l):
    for i, a in enumerate(l):
        for j, b in enumerate(l):
            if i != j and i != CurrencyType.NONE and j != CurrencyType.NONE:
                yield (a, b)
    yield (CurrencyType.NONE, CurrencyType.NONE)


class _PairType(BaseEnum):
    def __str__(self):
        return str(self.value[0].value) + '/' + str(self.value[1].value)

    def __hash__(self):
        return hash(str(self))

    @staticmethod
    @lru_cache(None)
    def from_string(first, second=''):
        if second:
            c1 = CurrencyType(first)
            c2 = CurrencyType(second)
            return PairType((c1, c2))
        first = first.strip().upper().replace('-', '/')
        if '/' not in first:
            for i in range(len(first)):
                if i == 0:
                    continue
                try:
                    c1 = CurrencyType(first[:i])
                    c2 = CurrencyType(first[i:])
                    return PairType.from_string(c1, c2)
                except ValueError:
                    continue
        first, second = first.split('/')
        c1 = CurrencyType(first)
        c2 = CurrencyType(second)
        return PairType((c1, c2))


PairType = _PairType('PairType', {(x[0].value + x[1].value if x[0] != CurrencyType.NONE else x[0].value): (x[0], x[1]) for x in _joiner(CurrencyType.__members__.values())})


class Side(BaseEnum):
    NONE = 'NONE'
    BUY = 'BUY'
    SELL = 'SELL'


class OptionSide(BaseEnum):
    NONE = 'NONE'
    CALL = 'CALL'
    PUT = 'PUT'


class OrderType(BaseEnum):
    NONE = 'NONE'
    MARKET = 'MARKET'
    LIMIT = 'LIMIT'


class OrderSubType(BaseEnum):
    NONE = 'NONE'
    POST_ONLY = 'POST_ONLY'
    FILL_OR_KILL = 'FILL_OR_KILL'
    # ALL_OR_NOTHING = 3


class TradeResult(BaseEnum):
    NONE = 'NONE'
    PENDING = 'PENDING'
    PARTIAL = 'PARTIAL'
    FILLED = 'FILLED'
    REJECTED = 'REJECTED'


class InstrumentType(BaseEnum):
    COIN = 'COIN'
    PAIR = 'PAIR'
    OPTION = 'OPTION'
    FUTURE = 'FUTURE'


class RiskReason(BaseEnum):
    NONE = 'NONE'
    PARTIAL = 'PARTIAL'
    FULL = 'FULL'



from datetime import datetime
from dataclasses import dataclass, asdict
from enum import Enum
from .enums import Side, \
    ExchangeType, \
    ExchangeType_to_string, \
    OptionSide, \
    CurrencyType, \
    PairType, \
    OrderType, \
    OrderSubType, \
    TickType, \
    TickType_to_string, \
    TradeResult, \
    InstrumentType, \
    RiskReason


@dataclass(init=False)
class Struct:
    def to_dict(self, serializable=False, str_timestamp=False, **kwargs) -> dict:
        ret = asdict(self)
        if serializable:
            for item in ret:
                if isinstance(ret[item], datetime):
                    if str_timestamp:
                        ret[item] = ret[item].strftime('%y-%m-%d %H:%M:%S')
                    else:
                        ret[item] = round(ret[item].timestamp())
                elif isinstance(ret[item], Struct) or isinstance(getattr(self, item), Struct):
                    ret[item] = getattr(self, item).to_dict(serializable, str_timestamp, **kwargs)
                elif isinstance(ret[item], Enum) or isinstance(getattr(self, item), Enum):
                    ret[item] = str(getattr(self, item))
                elif isinstance(ret[item], TickType):
                    ret[item] = TickType_to_string(ret[item])
                elif isinstance(ret[item], ExchangeType):
                    ret[item] = ExchangeType_to_string(ret[item])
                elif isinstance(ret[item], float):
                    if ((ret[item] >= float('inf')) is False) and \
                       ((ret[item] <= float('inf')) is False):
                        ret[item] = None
        return ret


@dataclass(init=False)
class Instrument(Struct):
    underlying: PairType
    type: InstrumentType = InstrumentType.PAIR

    def __init__(self,
                 underlying: PairType,
                 type: InstrumentType = InstrumentType.PAIR,
                 *args, **kwargs):
        self.underlying = underlying
        self.type = type

    @property
    def instrument(self):
        return self

    @property
    def currency_pair(self):
        return self.underlying

    def __eq__(self, other):
        return other.currency_pair == self.currency_pair

    def __str__(self):
        return str(self.underlying)

    def __repr__(self):
        return str(self.underlying)

    def __hash__(self):
        return hash(str(self.underlying))


@dataclass(init=False)
class Option(Instrument):
    underlying: Instrument
    expiration: datetime
    strike: float
    size: float
    side: OptionSide

    def __init__(self,
                 underlying: PairType,
                 expiration: datetime,
                 strike: float,
                 size: float,
                 side: OptionSide,
                 type: InstrumentType = InstrumentType.OPTION,
                 *args, **kwargs):
        self.underlying = underlying
        self.expiration = expiration
        self.strike = strike
        self.size = size
        self.side = side
        self.type = type
        super(Option, self).__init__(underlying=underlying,
                                     type=type,
                                     expiration=expiration,
                                     strike=strike,
                                     size=size,
                                     side=side)

    @property
    def instrument(self):
        return self.underlying

    @property
    def currency_pair(self):
        return self.underlying.currency_pair

    def __eq__(self, other):
        return (other.underlying == self.underlying) and \
               (other.currency_pair == self.currency_pair) and \
               (other.expiration == self.expiration) and \
               (other.strike == self.strike) and \
               (other.size == self.size)

    def __str__(self):
        # 180803C00187500-AAPL-CALL
        return self.expiration.strftime('%y%m%d') + \
            self.side.value[:1] + \
            '{0:10.2f}'.format(self.strike).replace(' ', '0') + '-' + \
            str(self.underlying) + '-' + \
            '{0:8.2f}'.format(self.size).replace(' ', '0')


@dataclass(init=False)
class Future(Instrument):
    underlying: Instrument
    expiration: datetime
    size: float

    def __init__(self,
                 underlying: PairType,
                 expiration: datetime,
                 size: float,
                 type: InstrumentType = InstrumentType.FUTURE,
                 *args, **kwargs):
        self.underlying = underlying
        self.expiration = expiration
        self.size = size
        self.type = type
        super(Future, self).__init__(underlying=underlying,
                                     type=type,
                                     expiration=expiration,
                                     size=size)

    @property
    def instrument(self):
        return self.underlying

    @property
    def currency_pair(self):
        return self.underlying.currency_pair

    def __eq__(self, other):
        return (other.underlying == self.underlying) and \
               (other.currency_pair == self.currency_pair) and \
               (other.expiration == self.expiration) and \
               (other.size == self.size)

    def __str__(self):
        # 180803C00187500-AAPL-CALL
        return f'<{self.expiration.strftime("%y%m%d")} - {self.underlying} -  ' + '{0:8.2f}'.format(self.size).replace(' ', '0') + '>'


@dataclass
class MarketData(Struct):
    # common
    time: datetime
    volume: float
    price: float
    type: TickType
    instrument: Instrument
    side: Side

    # maybe specific
    exchange: ExchangeType
    remaining: float = 0.0
    sequence: int = -1
    order_type: OrderType = OrderType.NONE
    order_id: str = ''

    def __eq__(self, other):
        return (self.price == other.price) and \
               (self.instrument == other.instrument) and \
               (self.side == other.side)

    def __str__(self):
        return f'<{self.instrument}-{self.volume}@{self.price}-{self.type}-{self.exchange}-{self.order_id}>'

    def __lt__(self, other):
        return self.price < other.price


@dataclass
class TradeRequest(Struct):
    side: Side
    exchange: ExchangeType

    volume: float
    price: float
    instrument: Instrument

    time: datetime

    order_type: OrderType
    order_sub_type: OrderSubType = OrderSubType.NONE
    risk_check: bool = False
    risk_reason: RiskReason = RiskReason.NONE
    strategy: object = None

    def __str__(self) -> str:
        return f'<{self.instrument}-{self.side}:{self.volume}@{self.price}-{self.order_type}-{self.exchange}>'


@dataclass
class TradeResponse(Struct):
    request: TradeRequest
    side: Side
    exchange: ExchangeType

    volume: float
    price: float
    instrument: Instrument

    time: datetime
    status: TradeResult
    order_id: str

    slippage: float = 0.0
    transaction_cost: float = 0.0
    remaining: float = 0.0
    strategy: object = None

    def __str__(self) -> str:
        return f'<{self.order_id}-{self.instrument}-{self.side}:{self.volume}@{self.price}-{self.status}-{self.exchange}>'


@dataclass
class Account(Struct):
    id: str
    currency: CurrencyType
    balance: float
    exchange: ExchangeType
    value: float
    asOf: datetime

    def __repr__(self) -> str:
        return f'<{self.id} - {self.currency} - {self.balance} - {self.value}>'



from functools import lru_cache
from typing import List
from .data_source import RestAPIDataSource
from .enums import TradingType, ExchangeType, ExchangeType_to_string, TradeResult
from .structs import TradeRequest, TradeResponse
from .utils import (get_keys_from_environment, str_to_side,
                    exchange_type_to_ccxt_client, tradereq_to_ccxt_order,
                    parse_date)


class OrderEntry(RestAPIDataSource):
    @lru_cache(None)
    def oe_client(self):
        options = self.options()
        if options.trading_type == TradingType.SANDBOX:
            key, secret, passphrase = get_keys_from_environment(ExchangeType_to_string(self.exchange()) + '_SANDBOX')
        else:
            key, secret, passphrase = get_keys_from_environment(ExchangeType_to_string(self.exchange()))

        return exchange_type_to_ccxt_client(self.exchange())({
            'apiKey': key,
            'secret': secret,
            'password': passphrase,
            'enableRateLimit': True,
            'rateLimit': 250
        })

    def _extract_fields(self, order, exchange):
        side = order.get('side', order.get('info', {}).get('side'))
        filled = float(order.get('filled') or order.get('info', {}).get('filled_size') or order.get('info', {}).get('executed_amount'))
        price = order.get('price') or order.get('info', {}).get('price')
        datetime = order.get('datetime') or order.get('info', {}).get('timestamp')
        status = order.get('status')
        remaining = float(order.get('remaining') or order.get('info', {}).get('remaining') or order.get('info', {}).get('remaining_amount'))
        cost = order.get('fee', {}).get('cost', 0.0)

        # FIXME handle rejected orders
        if status is None and exchange == ExchangeType.GEMINI:
            # gemini
            original = float(order.get('info', {}).get('original_amount', 0.0))
            is_cancelled = order.get('info', {}).get('is_cancelled', False)
            if is_cancelled:
                status = TradeResult.REJECTED
            if filled == original or remaining <= 0:
                status = TradeResult.FILLED
            elif remaining < original and remaining > 0:
                status = TradeResult.PARTIAL
            elif remaining == original:
                status = TradeResult.PENDING
        elif status in ('OPEN',):
            status = TradeResult.PENDING
        return side, filled, price, datetime, status, cost, remaining

    def buy(self, req: TradeRequest) -> TradeResponse:
        '''execute a buy order'''
        params = tradereq_to_ccxt_order(req)
        order = self.oe_client().create_order(**params)
        side, filled, price, datetime, status, cost, remaining = self._extract_fields(order, req.exchange)
        resp = TradeResponse(request=req,
                             side=str_to_side(side),
                             exchange=req.exchange,
                             volume=float(filled),
                             price=float(price),
                             instrument=req.instrument,
                             time=parse_date(datetime),
                             status=status,
                             order_id=order['id'],
                             slippage=float(price) - req.price,
                             transaction_cost=cost,
                             remaining=float(remaining))
        return resp

    def sell(self, req: TradeRequest) -> TradeResponse:
        '''execute a sell order'''
        params = tradereq_to_ccxt_order(req)
        order = self.oe_client().create_order(**params)
        side, filled, price, datetime, status, cost, remaining = self._extract_fields(order, req.exchange)
        resp = TradeResponse(request=req,
                             side=str_to_side(side),
                             exchange=req.exchange,
                             volume=float(filled),
                             price=float(price),
                             instrument=req.instrument,
                             time=parse_date(datetime),
                             status=status,
                             order_id=order['id'],
                             slippage=float(price) - req.price,
                             transaction_cost=cost,
                             remaining=float(remaining))
        return resp

    def cancel(self, resp: TradeResponse):
        self.oe_client().cancel_order(resp.order_id)

    def cancelAll(self, order_ids: List[str] = None):
        if order_ids:
            for order_id in order_ids:
                self.oe_client().cancel_order(order_id)
        else:
            self.oe_client().cancel_all_orders()



from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer
# from sqlalchemy import ForeignKey
# from sqlalchemy.orm import relationship

# Base sqlalchemy
Base = declarative_base()


class Record(Base):
    __tablename__ = 'records'
    id = Column(Integer, primary_key=True)



import tornado.gen
from ...enums import ExchangeType
from .base import HTTPHandler
from tornado.concurrent import run_on_executor
from perspective import PerspectiveHTTPMixin


class InstrumentsHandler(PerspectiveHTTPMixin, HTTPHandler):
    '''Server Handler
    Extends:
        HTTPHandler
    '''

    def initialize(self, trading_engine, psp_kwargs=None):
        self.te = trading_engine
        self.psp_kwargs = psp_kwargs or {}

    @run_on_executor
    def get_data(self, exchange, **psp_kwargs):
        if exchange:
            exchange = ExchangeType(exchange)
        msgs = [x.to_dict(True, True) for x in self.te.query.query_instruments(exchange)]
        super(InstrumentsHandler, self).loadData(data=msgs, **psp_kwargs)
        return super(InstrumentsHandler, self).getData()

    @tornado.web.authenticated
    @tornado.gen.coroutine
    def get(self):
        exchange = self.get_argument('exchange', '')
        dat = yield self.get_data(exchange, **self.psp_kwargs)
        self.write(dat)



# from abc import ABCMeta
from .structs import Instrument
from .enums import Side, TickType
# from heapq import heappush, heappop


class Order(object):
    def __init__(self, price, volume):
        self.price = price
        self.volume = volume

    def __lt__(self, other):
        return self.price < other.price


class Book(object):
    def __init__(self, instrument: Instrument):
        self._instrument = instrument
        self._bid = []
        self._bidd = {}

        self._ask = []
        self._askk = {}

    def push(self, order) -> None:
        price = round(order.price, 2)
        volume = round(order.volume, 4)

        if order.side == Side.BUY:
            if order.type == TickType.FILL or \
               order.type == TickType.CANCEL:
                if price not in self._bidd:
                    pass
                else:
                    self._bidd[price].volume -= volume
                    if self._bidd[price].volume < 1e-5:
                        self._bid.remove(self._bidd[price])
                        del self._bidd[price]

            elif order.type == TickType.CHANGE:
                if price not in self._bidd:
                    pass
                else:
                    self._bidd[price].volume -= volume
                    if self._bidd[price].volume < 1e-5:
                        self._bid.remove(self._bidd[price])
                        del self._bidd[price]

            elif order.type == TickType.OPEN:
                if order.price in self._bidd:
                    self._bidd[price].volume += volume
                else:
                    self._bid.append(Order(price, volume))
                    self._bidd[price] = self._bid[-1]
            else:
                self._bid.append(Order(price, volume))
                self._bidd[price] = self._bid[-1]
        else:
            if order.type == TickType.FILL or \
               order.type == TickType.CANCEL:
                if price not in self._askk:
                    pass
                else:
                    self._askk[price].volume -= volume
                    if self._askk[price].volume < 1e-5:
                        self._ask.remove(self._askk[price])
                        del self._askk[price]

            elif order.type == TickType.CHANGE:
                if price not in self._askk:
                    pass
                else:
                    self._askk[price].volume -= order.volume
                    if self._askk[price].volume < 1e-5:
                        self._ask.remove(self._askk[price])
                        del self._askk[price]

            elif order.type == TickType.OPEN:
                if price in self._askk:
                    self._askk[price].volume += volume
                else:
                    self._ask.append(Order(price, volume))
                    self._askk[price] = self._ask[-1]

            else:
                self._ask.append(Order(price, volume))
                self._askk[price] = self._ask[-1]

    def pop(self, order) -> None:
        pass

    def __str__(self) -> str:
        return str(self._instrument) + '->\n' + \
            'ask:\t' + '\n\t'.join(['%.1f\t@\t%.1f' % (x.volume, x.price) for x in sorted(self._ask, reverse=True)]) + \
            '\n\t=====================\n' + \
            'bid:\t' + '\n\t'.join(['%.1f\t@\t%.1f' % (x.volume, x.price) for x in sorted(self._bid, reverse=True)]) + '\n'

    def __repr__(self) -> str:
        return self.__str__()


# class OrderBook(metaclass=ABCMeta):
class OrderBook(object):
    '''OrderBook interface'''

    def __init__(self, instruments):
        self._ob = {instrument: Book(instrument) for instrument in instruments}

    def preload(self, orders) -> None:
        pass

    def push(self, order) -> None:
        self._ob[order.instrument].push(order)

    def tob(self) -> list:
        return self._ob

    def __str__(self) -> str:
        return '\n\n'.join([str(self._ob[b]) for b in self._ob])

    def __repr__(self) -> str:
        return self.__str__()




class TestDataSource:
    def setup(self):
        pass

    def teardown(self):
        pass

    @classmethod
    def setup_class(cls):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test_StreamingDataSource(self):
        from ..data_source import StreamingDataSource
        from ..callback import Callback
        from ..enums import TickType

        try:
            x = StreamingDataSource()
            assert False
        except:
            pass

        class Test(StreamingDataSource):
            def run(self, engine):
                pass

            def close(self):
                pass

            def receive(self):
                pass

            def seqnum(self, num):
                pass

            def tickToData(self):
                pass

        class TestCB(Callback):
            def onTrade(self):
                pass

            def onOpen(self):
                pass

            def onFill(self):
                pass

            def onCancel(self):
                pass

            def onChange(self):
                pass

            def onError(self):
                pass

            def onAnalyze(self):
                pass

            def onHalt(self, data):
                pass

            def onContinue(self, data):
                pass

        try:
            t = Test()

            t.registerCallback(TestCB())
            assert t._callbacks
            assert len(t._callbacks) == 9
            assert len(t._callbacks[TickType.ERROR]) == 1

        except Exception as e:
            print(e)
            assert False



import pandas as pd


class TestBacktest:
    def setup(self):
        from ..config import BacktestConfig
        self.config = BacktestConfig()
        df = pd.DataFrame([{'volume': 100.0,
                            'close': 1.0,
                            'timestamp': '1558296780000',
                            'exchange': 'GEMINI',
                            'pair': 'USDBTC'
                            }])
        df['timestamp'] = pd.to_datetime(df['timestamp'], unit='ms')
        df.set_index(['timestamp', 'pair'], inplace=True)
        self.test_line = df.iloc[0]

    def teardown(self):
        pass
        # teardown() after each test method

    @classmethod
    def setup_class(cls):
        from ..callback import Callback

        class CallbackTester(Callback):
            def __init__(self):
                self._onTrade = False
                self._onReceived = False
                self._onOpen = False
                self._onFill = False
                self._onCancel = False
                self._onChange = False
                self._onError = False
                self._onAnalyze = False
                self._onHalt = False
                self._onContinue = False

            def onTrade(self, data):
                self._onTrade = True

            def onReceived(self, data):
                self._onReceived = True

            def onOpen(self, data):
                self._onOpen = True

            def onFill(self, data):
                self._onFill = True

            def onCancel(self, data):
                self._onCancel = True

            def onChange(self, data):
                self._onChange = True

            def onError(self, data):
                self._onError = True

            def onAnalyze(self, data):
                self._onAnalyze = True

            def onHalt(self, data):
                self._onHalt = True

            def onContinue(self, data):
                self._onContinue = True

        cls.demo_callback = CallbackTester

    @classmethod
    def teardown_class(cls):
        pass
        # teardown_class() after any methods in this class

    def test_init(self):
        from ..backtest import Backtest
        b = Backtest(self.config)
        assert b

    def test_receive(self):
        from ..backtest import Backtest, line_to_data

        b = Backtest(self.config)
        cb = self.demo_callback()

        b.registerCallback(cb)
        b.receive(line_to_data(self.test_line))
        assert cb._onTrade



from mock import patch


class TestMain:
    def setup(self):
        pass
        # setup() before each test method

    def teardown(self):
        pass
        # teardown() after each test method

    @classmethod
    def setup_class(cls):
        pass
        # setup_class() before any methods in this class

    @classmethod
    def teardown_class(cls):
        pass
        # teardown_class() after any methods in this class

    def test_parse_command_line(self):
        from ..main import parse_command_line_config
        from ..enums import TradingType

        argv = ['', '--live', '--exchanges=coinbase']
        ret = parse_command_line_config(argv)
        assert ret.type == TradingType.LIVE

        argv = ['', '--sandbox', '--exchanges=coinbase']
        ret = parse_command_line_config(argv)
        assert ret.type == TradingType.SANDBOX

        argv = ['', '--backtest', '--exchanges=coinbase']
        ret = parse_command_line_config(argv)
        assert ret.type == TradingType.BACKTEST

        argv = ['', '--live', '--print', '--exchanges=coinbase']
        ret = parse_command_line_config(argv)
        assert ret.print

    def test_main(self):
        from ..main import main
        with patch('aat.main.TradingEngine'), \
             patch('aat.strategies.buy_and_hold.BuyAndHoldStrategy'):
            main(['', '--live', '--exchanges=coinbase'])



from mock import MagicMock
from datetime import datetime


class TestRisk:
    def setup(self):
        from ..config import RiskConfig
        from ..risk import Risk
        from ..enums import ExchangeType, CurrencyType
        from ..structs import Account

        rc = RiskConfig()
        rc.max_risk = 100.0
        rc.max_drawdown = 100.0
        rc.total_funds = 100.0

        ex = {ExchangeType.COINBASE: MagicMock()}
        accounts = {CurrencyType.BTC: {ExchangeType.COINBASE: Account(id='1',
                                                                      currency=CurrencyType.BTC,
                                                                      balance=1.0,
                                                                      exchange=MagicMock(),
                                                                      value=-1,
                                                                      asOf=datetime.now())},
                    CurrencyType.USD: {ExchangeType.COINBASE: Account(id='2',
                                                                      currency=CurrencyType.USD,
                                                                      balance=1.0,
                                                                      exchange=MagicMock(),
                                                                      value=-1,
                                                                      asOf=datetime.now())}}
        ex[ExchangeType.COINBASE].accounts.return_value = accounts

        self.risk = Risk(rc, ex, accounts)

        # setup() before each test method

    def test_construct_reponse(self):
        pass

    def test_request(self):
        from ..structs import TradeRequest, Instrument
        from ..enums import Side, PairType, OrderType, ExchangeType

        req = TradeRequest(side=Side.BUY, instrument=Instrument(underlying=PairType.BTCUSD), order_type=OrderType.MARKET, volume=100.0, price=1.0, exchange=ExchangeType.COINBASE, time=datetime.now())
        resp = self.risk.request(req)
        resp = self.risk.requestBuy(req)

        # FIXME
        # assert resp.risk_check == True
        # assert resp.volume == 100.0
        # assert self.risk.outstanding == 0.0
        # assert self.risk.max_running_outstanding == 0.0
        # print(self.risk.max_running_outstanding_incr)
        # assert self.risk.max_running_outstanding_incr == []

        # req = TradeRequest(side=Side.BUY, volume=100.0, price=1.0)
        # resp = self.risk.request(req)

        # assert resp.risk_check == True

    def test_request2(self):
        from ..structs import TradeRequest, Instrument
        from ..enums import Side, PairType, OrderType, ExchangeType

        req = TradeRequest(side=Side.SELL, instrument=Instrument(underlying=PairType.BTCUSD), order_type=OrderType.MARKET, volume=50.0, price=1.0, exchange=ExchangeType.COINBASE, time=datetime.now())
        resp = self.risk.request(req)
        resp = self.risk.requestSell(req)

        # FIXME
        # assert resp.risk_check == True
        # assert resp.volume == 50.0
        # assert self.risk.outstanding == 50.0
        # assert self.risk.max_running_outstanding == 50.0
        # assert self.risk.max_running_outstanding_incr == [50.0]

        # req = TradeRequest(side=Side.BUY, volume=100.0, price=1.0)
        # resp = self.risk.request(req)

        # assert resp.risk_check == True
        # assert resp.volume == 50.0
        # assert self.risk.outstanding == 100.0
        # assert self.risk.max_running_outstanding == 100.0
        # assert self.risk.max_running_outstanding_incr == [50.0, 100.0]

        # req = TradeRequest(side=Side.BUY, volume=100.0, price=1.0)
        # resp = self.risk.request(req)

        # assert resp.risk_check == False

    def test_update(self):
        from ..structs import TradeRequest, TradeResponse, Instrument
        from ..enums import Side, PairType, OrderType, ExchangeType, TradeResult
        req = TradeRequest(side=Side.SELL, instrument=Instrument(underlying=PairType.BTCUSD), order_type=OrderType.MARKET, volume=50.0, price=1.0, exchange=ExchangeType.COINBASE, time=datetime.now())
        resp = TradeResponse(side=Side.SELL, instrument=Instrument(underlying=PairType.BTCUSD), request=req, order_id='1', volume=50.0, price=1.0, exchange=ExchangeType.COINBASE, time=datetime.now(), status=TradeResult.FILLED)
        self.risk.update(resp)
        assert self.risk.outstanding == -50

    def test_cancel(self):
        from ..structs import TradeRequest, TradeResponse, Instrument
        from ..enums import Side, PairType, OrderType, ExchangeType, TradeResult
        req = TradeRequest(side=Side.SELL, instrument=Instrument(underlying=PairType.BTCUSD), order_type=OrderType.MARKET, volume=50.0, price=1.0, exchange=ExchangeType.COINBASE, time=datetime.now())
        resp = TradeResponse(side=Side.SELL, instrument=Instrument(underlying=PairType.BTCUSD), request=req, order_id='1', volume=50.0, price=1.0, exchange=ExchangeType.COINBASE, time=datetime.now(), status=TradeResult.FILLED)
        self.risk.cancel(resp)
        assert self.risk.outstanding == 50



import tornado.ioloop
import tornado.web
import tornado.websocket
from concurrent.futures import ThreadPoolExecutor
from jinja2 import Environment, FileSystemLoader, TemplateNotFound


class HTTPHandler(tornado.web.RequestHandler):
    '''Just a default handler'''
    executor = ThreadPoolExecutor(16)

    def get_current_user(self):
        return self.get_secure_cookie("token")

    def initialize(self, *args, **kwargs):
        '''Initialize the server competition registry handler

        This handler is responsible for managing competition
        registration.

        Arguments:
            competitions {dict} -- a reference to the server's dictionary of competitions
        '''
        super(HTTPHandler, self).initialize(*args, **kwargs)

    def render_template(self, template, **kwargs):
        if hasattr(self, 'template_dirs'):
            template_dirs = self.template_dirs
        else:
            template_dirs = []

        if self.settings.get('template_path', ''):
            template_dirs.append(
                self.settings["template_path"]
            )
        env = Environment(loader=FileSystemLoader(template_dirs))

        try:
            template = env.get_template(template)
        except TemplateNotFound:
            raise TemplateNotFound(template)

        kwargs['current_user'] = self.current_user.decode('utf-8') if self.current_user else ''
        content = template.render(**kwargs)
        return content


class WebSocketHandler(tornado.websocket.WebSocketHandler):
    '''Just a default handler'''

    def initialize(self, *args, **kwargs):
        '''Initialize the server competition registry handler

        This handler is responsible for managing competition
        registration.

        Arguments:
            competitions {dict} -- a reference to the server's dictionary of competitions
        '''
        super(WebSocketHandler, self).initialize(*args, **kwargs)



import tornado
from .base import HTTPHandler
from ...logging import log


class LoginHandler(HTTPHandler):
    def get(self):
        '''Get the login page'''
        if self.current_user:
            self.redirect('/')
        else:
            log.critical(f'\n**********\nLogin code: {self.application.settings["login_code"]}\n**********')

            user = self.get_argument('code', '')
            if user == self.application.settings['login_code']:
                self.set_secure_cookie("token", user)
            else:
                self.redirect('/login')  # 404 in raw AAT

    def post(self):
        '''Login'''
        user = self.get_argument('code', '')
        if user == self.application.settings['login_code']:
            self.set_secure_cookie("token", user)
            self.redirect('/')
        else:
            self.redirect('/login')  # 404 in raw AAT


class LogoutHandler(HTTPHandler):
    @tornado.web.authenticated
    def get(self):
        '''Get the logout page'''
        if not self.current_user:
            self.redirect('/login')
        else:
            self.redirect('/logout')  # 404 in raw AAT

    def post(self):
        self.clear_cookie("user")
        self.redirect('/')



from .common import generateTornadoApplication
from ....ui.handlers.base import HTTPHandler
from mock import MagicMock


class TestBase:
    def setup(self):
        self.app, self.login_code = generateTornadoApplication()

    def test_validate(self):
        req = MagicMock()
        req.body = ''
        x = HTTPHandler(self.app, req)
        x._transforms = []
        x.render_template('404.html')



class TestCallback:
    def setup(self):
        pass
        # setup() before each test method

    def teardown(self):
        pass
        # teardown() after each test method

    @classmethod
    def setup_class(cls):
        # setup_class() before any methods in this class
        pass

    @classmethod
    def teardown_class(cls):
        # teardown_class() after any methods in this class
        pass

    def test_null_callback(self):
        from ..callback import NullCallback

        nc = NullCallback()
        assert nc.onTrade(None) is None
        assert nc.onOpen(None) is None
        assert nc.onFill(None) is None
        assert nc.onCancel(None) is None
        assert nc.onChange(None) is None
        assert nc.onError(None) is None
        assert nc.onAnalyze(None) is None
        assert nc.onHalt(None) is None
        assert nc.onContinue(None) is None

    def test_print_callback(self):
        from ..callback import Print
        pc = Print(onError=False)
        assert pc.onError == False
        assert pc.onTrade('test-print_onTrade') == None



import json
from functools import lru_cache
from datetime import datetime
from ..enums import PairType, TickType, TickType_from_string
from ..exchange import Exchange
from ..structs import MarketData, Instrument
from ..utils import parse_date, str_to_side, str_to_order_type


class CoinbaseExchange(Exchange):
    @lru_cache(None)
    def subscription(self):
        return [json.dumps({"type": "subscribe", "product_id": x.value[0].value + '-' + x.value[1].value}) for x in self.options().currency_pairs]

    @lru_cache(None)
    def heartbeat(self):
        return json.dumps({"type": "heartbeat", "on": True})

    def tickToData(self, jsn: dict) -> MarketData:
        '''convert a jsn tick off the websocket to a MarketData struct'''
        if jsn.get('type') == 'received':
            return

        s = jsn.get('type').upper()
        reason = jsn.get('reason', '').upper()
        if s == 'MATCH' or (s == 'DONE' and reason == 'FILLED'):
            typ = TickType.TRADE
        elif s in ('OPEN', 'DONE', 'CHANGE', 'HEARTBEAT'):
            if reason == 'CANCELED':
                typ = TickType.CANCEL
            elif s == 'DONE':
                typ = TickType.FILL
            else:
                typ = TickType_from_string(s.upper())
        else:
            typ = TickType.ERROR

        order_id = jsn.get('order_id', jsn.get('maker_order_id', ''))
        time = parse_date(jsn.get('time')) if jsn.get('time') else datetime.now()

        if typ in (TickType.CANCEL, TickType.OPEN):
            volume = float(jsn.get('remaining_size', 'nan'))
        else:
            volume = float(jsn.get('size', 'nan'))
        price = float(jsn.get('price', 'nan'))

        currency_pair = PairType.from_string(jsn.get('product_id')) if typ != TickType.ERROR else PairType.NONE

        instrument = Instrument(underlying=currency_pair)

        order_type = str_to_order_type(jsn.get('order_type', ''))
        side = str_to_side(jsn.get('side', ''))
        remaining_volume = float(jsn.get('remaining_size', 0.0))

        sequence = int(jsn.get('sequence', -1))
        ret = MarketData(order_id=order_id,
                         time=time,
                         volume=volume,
                         price=price,
                         type=typ,
                         instrument=instrument,
                         remaining=remaining_volume,
                         side=side,
                         exchange=self.exchange(),
                         order_type=order_type,
                         sequence=sequence)
        return ret



import os
from mock import patch


class TestOrderEntry:
    def setup(self):
        # setup() before each test method
        from aat.config import ExchangeConfig
        from aat.enums import TradingType, ExchangeType, PairType

        os.environ['GEMINI_API_KEY'] = 'test'
        os.environ['GEMINI_API_SECRET'] = 'test'
        os.environ['GEMINI_API_PASS'] = 'test'

        self.ec = ExchangeConfig(exchange_types=[ExchangeType.GEMINI],
                                 trading_type=TradingType.LIVE,
                                 currency_pairs=[PairType.BTCUSD])



import operator
# from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from functools import reduce
from typing import List, Dict
from .enums import TradeResult, ExchangeType, PairType, CurrencyType, TradingType, Side
from .exceptions import QueryException, AATException
from .execution import Execution
from .logging import log
from .risk import Risk
from .strategy import TradingStrategy
from .structs import Instrument, MarketData, TradeRequest, TradeResponse
from .utils import iterate_accounts, pnl_helper, findpath


class QueryEngine(object):
    def __init__(self,
                 trading_type: TradingType = None,
                 exchanges: List[ExchangeType] = None,
                 pairs: List[PairType] = None,
                 instruments: List[Instrument] = None,
                 accounts=None,
                 risk: Risk = None,
                 execution: Execution = None):
        # self._executor = ThreadPoolExecutor(16)
        self._all = []
        self._trading_type = trading_type

        self._accounts = accounts

        # public
        self.positions_value = [[datetime.now(), 0.0, 0.0, 0.0]]
        self.portfolio_value = [[datetime.now(), risk.total_funds]]
        self.positions = {}
        self.pending = {}

        self._trades = []
        self._trades_by_instrument = {}

        self._pairs = pairs

        # public
        self.instruments = instruments
        self.exchanges = exchanges

        self._last_price_by_asset_and_exchange = {}

        self._trade_reqs = []
        self._trade_resps = []
        self._trade_reqs_by_instrument = {}
        self._trade_resps_by_instrument = {}

        # public
        self.strategies = []

        self._risk = risk
        self._execution = execution

    def registerStrategy(self, strat: TradingStrategy):
        self.strategies.append(strat)

    def query_instruments(self, exchange=None) -> List[PairType]:
        '''get list of all instruments available on all exchanges'''
        return self._instruments[exchange] if exchange else reduce(operator.concat, self._instruments.values())

    def query_exchanges(self) -> List[Dict]:
        '''get list of exchanges and their available instruments'''
        return [{'exchange': name, 'instruments': self.query_instruments(name)} for name, ex in self._exchanges.items()]

    def _paginate(self, instrument: Instrument, lst: list, lst_sub: list, page: int = 1) -> list:
        '''paginate a data request'''
        page, from_, to_ = (page, -1 * page * 100, -1 * (page - 1) * 100) if page is not None else (0, 0, -1)

        if instrument:
            return lst_sub[instrument][from_:to_] if page > 1 else lst_sub[instrument][from_:]
        return lst[from_:to_] if page > 1 else lst[from_:]

    def query_lastpriceall(self) -> List[MarketData]:
        '''get last price of all assets'''
        return [m for exs in self._last_price_by_asset_and_exchange.values() for m in exs.values()]

    def query_lastprice(self, instrument: Instrument, exchange: ExchangeType = None) -> MarketData:
        '''get last price of asset, potentially on a given exchange'''
        if instrument not in self._last_price_by_asset_and_exchange:
            raise QueryException('Not found!')
        if exchange:
            if exchange not in self._last_price_by_asset_and_exchange[instrument]:
                raise QueryException('Not found!')
            return self._last_price_by_asset_and_exchange[instrument][exchange]
        if "ANY" not in self._last_price_by_asset_and_exchange[instrument]:
            raise QueryException('Not found!')
        return self._last_price_by_asset_and_exchange[instrument]["ANY"]

    def query_trades(self, instrument: Instrument = None, page: int = 1) -> List[MarketData]:
        '''get trades for instrument'''
        return self._paginate(instrument,
                              self._trades,
                              self._trades_by_instrument,
                              page)

    def query_tradereqs(self, instrument: Instrument = None, page: int = 1) -> List[TradeRequest]:
        '''get trade requests for an instrument'''
        return self._paginate(instrument, self._trade_reqs, self._trade_reqs_by_instrument, page)

    def query_traderesps(self, instrument: Instrument = None, page: int = 1) -> List[TradeResponse]:
        '''get trade responses for an instrument'''
        return self._paginate(instrument, self._trade_resps, self._trade_resps_by_instrument, page)

    def newPending(self, resp: TradeResponse) -> None:
        self.pending[resp.order_id] = resp

    def push_tradereq(self, req: TradeRequest) -> None:
        '''append trade request to list'''
        self._trade_reqs.append(req)
        if req.instrument not in self._trade_reqs_by_instrument:
            self._trade_reqs_by_instrument[req.instrument] = []
        self._trade_reqs_by_instrument[req.instrument].append(req)

    def push_traderesp(self, resp: TradeResponse) -> None:
        '''append trade response to list'''
        if resp.status in (TradeResult.REJECTED, TradeResult.PENDING, TradeResult.NONE):
            return
        self._trade_resps.append(resp)
        if resp.instrument not in self._trade_resps_by_instrument:
            self._trade_resps_by_instrument[resp.instrument] = []
        self._trade_resps_by_instrument[resp.instrument].append(resp)

    def onTrade(self, data: MarketData) -> None:
        '''process market data on trade'''

        # append to list of all data so far
        self._all.append(data)

        # append to list of all trades so far
        self._trades.append(data)

        # if not tracking, initialize list of trades
        if data.instrument not in self._trades_by_instrument:
            self._trades_by_instrument[data.instrument] = []
        # add data to list
        self._trades_by_instrument[data.instrument].append(data)

        # if not tracking by exchange
        if data.instrument not in self._last_price_by_asset_and_exchange:
            self._last_price_by_asset_and_exchange[data.instrument] = {}
        # add data to
        self._last_price_by_asset_and_exchange[data.instrument][data.exchange] = data

        # set data to be last on ANY exchange
        self._last_price_by_asset_and_exchange[data.instrument]['ANY'] = data

        # if any pending orders for this trade
        if data.order_id in self.pending.keys():
            # grab previous pending response
            resp = self.pending[data.order_id]

            # if they're equal, ignore remaining
            resp.volume = resp.remaining - data.remaining if resp != data else resp.volume
            resp.remaining = data.remaining

            # call on fill to strat
            if resp.strategy:
                resp.strategy.onFill(resp)
            else:
                raise Exception('Response has no corresponding strategy!')

            # update account holdings
            self.updateAccounts(resp)

            # update porfolio risk
            self._risk.update(resp)

            # delete pending order if filled
            if data.remaining <= 0:
                del self.pending[data.order_id]
        else:
            # price only
            if data.instrument in self.positions:
                self.positions[data.instrument].price(data.price)

        # recalculate value of positions
        self._recalculate_positions(data)

        # recalculate value of portfolio
        self._recalculate_portfolio(data)

    def onFill(self, resp: TradeResponse) -> None:
        if self._trading_type not in (TradingType.BACKTEST, TradingType.SIMULATION):
            # only used during offline trading
            raise QueryException('Must derive fills from market data!')

        # ignore different responses
        if resp.volume <= 0 or resp.status != TradeResult.FILLED:
            # skip
            return
        resp.remaining = 0.0
        self.onTrade(resp)

    def onCancel(self, data: MarketData) -> None:
        # if pending order
        if data.order_id in self.pending.keys():
            # grab response
            resp = self.pending[data.order_id]

            # delete from tracking
            del self.pending[data.order_id]

            # tell risk
            self._risk.cancel(resp)

    def updateAccounts(self, resp: TradeResponse = None) -> None:
        '''update the holdings and spot value of accounts'''

        # if reacting to a trade response
        if resp:

            # get left and right legs
            account_left = self._accounts[resp.instrument.underlying.value[0]][resp.exchange]
            account_right = self._accounts[resp.instrument.underlying.value[1]][resp.exchange]

            if resp.side == Side.BUY:
                # if buy
                # 5 BTCUSD @ $2 -> from_btc += volume, to_usd -= price*volume
                account_left.balance += resp.volume
                account_right.balance -= resp.volume * resp.price
            else:
                # if sell
                # 5 BTCUSD @ $2 -> from_btc -= volume, to_usd += price*volume
                account_left.balance -= resp.volume
                account_right.balance += resp.volume * resp.price
            # iterate through these
            accounts = (account_left, account_right)
        else:
            # WARNING
            # don't update value on every tick
            log.warn('Iterating through ALL accounts')
            accounts = iterate_accounts(self.accounts)

        # update account values
        for account in accounts:
            log.info(f'Updating value of account {account}')
            if account.currency == CurrencyType.USD:
                # if holding USD, add value
                account.value = account.balance
            else:
                # calculate USD value
                price = self._last_price_by_asset_and_exchange[resp.instrument][resp.exchange].price
                account.value = account.balance * price
            log.info(f'New value: {account}')

    def update_positions(self, resp: TradeResponse) -> None:
        '''update positions based on a trade'''
        if resp.status in (TradeResult.REJECTED, TradeResult.PENDING, TradeResult.NONE):
            '''ignore these as they don't affect what we hold'''
            return

        if resp.instrument not in self.positions:
            self.positions[resp.instrument] = pnl_helper()

        self.positions[resp.instrument].exec(resp.volume, resp.price, resp.side)

    def _recalculate_positions(self, data: MarketData) -> None:
        '''recalculate the market value of active positions'''
        # ignore if not an active position
        if data.instrument not in self.positions:
            return

        volume = self.positions[data.instrument]._volume
        # instrument = str(data.instrument)
        _ = str(data.instrument)
        # value = self.positions[data.instrument]._avg_price * volume
        _ = self.positions[data.instrument]._avg_price * volume
        pnl = self.positions[data.instrument]._pnl

        # get PnL numbers
        unrealized = sum(x._pnl for x in self.positions.values())
        realized = sum(x._realized for x in self.positions.values())
        pnl = sum(x._pnl + x._realized for x in self.positions.values())
        self.portfolio_value.append([data.time, self.portfolio_value[-1][1], unrealized, realized, pnl])

    def _recalculate_portfolio(self, data: MarketData) -> None:
        '''recalculate the market value of all accounts'''
        if data.instrument.underlying.value[0] == CurrencyType.USD:
            left_ret = 1
        else:
            left_ret = self._ticker(data.instrument.underlying.value[0], exchange=self.exchanges[data.exchange])

        if data.instrument.underlying.value[1] == CurrencyType.USD:
            right_ret = 1
        else:
            right_ret = self._ticker(data.instrument.underlying.value[1], exchange=self.exchanges[data.exchange])

        print(left_ret, right_ret)
        self.portfolio_value.append([data.time, self.portfolio_value[-1][-1]])

    def _ticker(self, currency: CurrencyType, exchange):
        inst = Instrument(underlying=PairType.from_string(currency.value + '/USD'))

        if inst in exchange.markets():
            return self.query_lastprice(inst)
        else:
            for stable in ('USD', 'USDC', 'USDT'):
                try:
                    inst = Instrument(underlying=PairType.from_string(currency.value + '/' + stable))
                    inst1, inst2, i1_inverted, i2_inverted = findpath(inst, exchange.markets())
                    broken = False
                except (AATException, ValueError):
                    broken = True
            if broken:
                return None

            inst1_t = self.query_lastprice(instrument=inst1).price
            inst2_t = self.query_lastprice(instrument=inst2).price
            if i1_inverted:
                inst1_t = 1.0 / inst1_t
            if i2_inverted:
                inst2_t = 1.0 / inst2_t
            px = inst1_t * inst2_t
            return px



from ..strategy import TradingStrategy
from ..structs import MarketData, TradeResponse
from ..logging import log


class DataCaptureStrategy(TradingStrategy):
    def __init__(self, filename, *args, **kwargs) -> None:
        super(DataCaptureStrategy, self).__init__(*args, **kwargs)
        self.filename = filename

    def onTrade(self, data: MarketData) -> bool:
        log.info(data)

    def onError(self, e) -> None:
        log.critical(e, type(e))

    def onExit(self) -> None:
        self.cancelAll(lambda *args: True)

    def onChange(self, data: MarketData) -> None:
        log.info(data)

    def onFill(self, res: TradeResponse) -> None:
        log.info(res)

    def onCancel(self, data: MarketData) -> None:
        log.info(data)

    def onOpen(self, data: MarketData) -> None:
        log.info(data)

    def slippage(self, resp: TradeResponse) -> TradeResponse:
        return resp

    def transactionCost(self, resp: TradeResponse) -> TradeResponse:
        return resp

    def onAnalyze(self, engine) -> None:
        pass



from .common import generateTornadoApplication
from ....ui.handlers.exchanges import ExchangesHandler
from mock import MagicMock, patch


class TestAccounts:
    def setup(self):
        self.app, self.login_code = generateTornadoApplication()

    def test_AccountsHandler(self):
        req = MagicMock()
        req.body = ''
        x = ExchangesHandler(self.app, req, trading_engine=MagicMock(), psp_kwargs={})

        x.te.exchanges = MagicMock()
        x.te.exchanges.return_value.values.return_value = [MagicMock()]
        x.te.exchanges.return_value.values.return_value[0].accounts.return_value = [MagicMock()]
        x.te.exchanges.return_value.values.return_value[0].accounts.return_value[0].to_dict.return_value = {'test': 1}

        x._transforms = []
        with patch.object(x, 'get_secure_cookie') as mget:
            mget.return_value = self.login_code
            x.get()



import pytest
import random
from datetime import datetime
from aat.order_book import OrderBook
from aat.structs import MarketData, Instrument
from aat.enums import Side, \
                      OptionSide, \
                      CurrencyType, \
                      PairType, \
                      OrderType, \
                      OrderSubType, \
                      TickType, \
                      TradeResult, \
                      InstrumentType, \
                      ExchangeType


def generateInstruments(pairs):
    return [Instrument(underlying=pair) for pair in pairs]


def initialMarketData(count, instruments=None):
    instruments = instruments or generateInstruments([PairType.BTCUSD, PairType.ETHUSD])
    ret = []
    for _ in range(count):
        side = random.choice([Side.BUY, Side.SELL])
        volume = random.randrange(0, 20) / 10

        if side == Side.BUY:
            price = random.randrange(500, 550) / 10

        if side == Side.SELL:
            price = random.randrange(551, 600) / 10

        tick = random.choice([TickType.OPEN])
        instrument = random.choice(instruments)

        remaining = random.choice([0.0, random.randrange(0, 100)/10])
        sequence = -1
        order_type = OrderType.NONE

        m = MarketData(time=datetime.now(),
                       volume=volume,
                       price=price,
                       type=tick,
                       instrument=instrument,
                       side=side,
                       remaining=remaining,
                       sequence=sequence,
                       exchange=ExchangeType.COINBASE,
                       order_type=order_type)

        ret.append(m)
    return ret


def generateMarketData(count, instruments=None):
    instruments = instruments or generateInstruments([PairType.BTCUSD, PairType.ETHUSD])
    ret = []
    for _ in range(count):
        side = random.choice([Side.BUY, Side.SELL])

        volume = random.randrange(0, 20) / 10

        if side == Side.BUY:
            price = random.randrange(500, 550) / 10

        if side == Side.SELL:
            price = random.randrange(551, 600) / 10

        tick = random.choice([TickType.OPEN,
                              TickType.DONE,
                              TickType.CHANGE])
        instrument = random.choice(instruments)

        remaining = random.choice([0.0, random.randrange(0, 100)/10])

        sequence = -1
        order_type = OrderType.NONE

        m = MarketData(time=datetime.now(),
                       volume=volume,
                       price=price,
                       type=tick,
                       instrument=instrument,
                       side=side,
                       remaining=remaining,
                       sequence=sequence,
                       exchange=ExchangeType.COINBASE,
                       order_type=order_type)

        ret.append(m)

    return ret


class TestDataSource:
    def test_order_book(self):
        pairs = [PairType.BTCUSD, PairType.ETHUSD]
        instruments = generateInstruments(pairs)

        ob = OrderBook(instruments)

        for item in initialMarketData(50, instruments):
            ob.push(item)

        print(str(ob))

    @pytest.mark.skip(reason="no way of currently testing this")
    def test_order_book_sequence(self):
        pairs = [PairType.BTCUSD]
        instruments = generateInstruments(pairs)
        ob = OrderBook(instruments)

        p = 0.0
        while p < 2.1:
            m = MarketData(time=datetime.now(),
                           volume=1.0,
                           price=p,
                           type=TickType.OPEN,
                           instrument=instruments[0],
                           side=Side.BUY,
                           remaining=0.0,
                           sequence=-1,
                           exchange=ExchangeType.COINBASE,
                           order_type=OrderType.NONE)
            if p > 1.0:
                m.side = Side.SELL
            ob.push(m)
            p += .1

        print(ob)

        p = 0.0
        while p < 2.1:
            if round(p * 10) % 2 < 1:
                m = MarketData(time=datetime.now(),
                               volume=1.0,
                               price=p,
                               type=TickType.OPEN,
                               instrument=instruments[0],
                               side=Side.BUY,
                               remaining=0.0,
                               sequence=-1,
                               exchange=ExchangeType.COINBASE,
                               order_type=OrderType.NONE)
                if p > 1.0:
                    m.side = Side.SELL
                ob.push(m)
            p += .1

        print(ob)

        p = 0.0
        while p < 2.1:
            m = MarketData(time=datetime.now(),
                           volume=1.0,
                           price=p,
                           type=TickType.CANCELLED,
                           instrument=instruments[0],
                           side=Side.BUY,
                           remaining=0.0,
                           sequence=-1,
                           exchange=ExchangeType.COINBASE,
                           order_type=OrderType.NONE)
            if p > 1.0:
                m.side = Side.SELL

            if round(p * 10) % 2 < 1:
                m.type = TickType.FILLED
            ob.push(m)
            p += .1

        print(ob)

        print(ob._ob[instruments[0]]._bidd.keys())
        print(ob._ob[instruments[0]]._bid)
        # TODO check floating point error
        assert ob._ob[instruments[0]]._bidd[0.0] == 1.0
        assert ob._ob[instruments[0]]._bidd[0.2] == 1.0
        assert ob._ob[instruments[0]]._bidd[0.4] == 1.0
        assert ob._ob[instruments[0]]._bidd[0.6] == 1.0
        assert ob._ob[instruments[0]]._bidd[0.8] == 1.0
        assert ob._ob[instruments[0]]._bidd[1.0] == 1.0
        assert ob._ob[instruments[0]]._askk[1.2] == 1.0
        assert ob._ob[instruments[0]]._askk[1.4] == 1.0
        assert ob._ob[instruments[0]]._askk[1.6] == 1.0
        assert ob._ob[instruments[0]]._askk[1.8] == 1.0
        assert ob._ob[instruments[0]]._askk[1.8] == 1.0



import tornado.gen
from .base import HTTPHandler
from tornado.concurrent import run_on_executor
from perspective import PerspectiveHTTPMixin


class StrategiesHandler(PerspectiveHTTPMixin, HTTPHandler):
    '''Server Handler
    Extends:
        HTTPHandler
    '''

    def initialize(self, trading_engine, psp_kwargs=None):
        self.te = trading_engine
        self.psp_kwargs = psp_kwargs or {}

    @run_on_executor
    def get_data(self, **psp_kwargs):
        dat = [s.to_dict() for s in self.te.strategies()]
        super(StrategiesHandler, self).loadData(data=dat, **psp_kwargs)
        return super(StrategiesHandler, self).getData()

    @tornado.web.authenticated
    @tornado.gen.coroutine
    def get(self):
        dat = yield self.get_data(**self.psp_kwargs)
        self.write(dat)



class TestEnums:
    def setup(self):
        pass
        # setup() before each test method

    def teardown(self):
        pass
        # teardown() after each test method

    @classmethod
    def setup_class(cls):
        pass
        # setup_class() before any methods in this class

    @classmethod
    def teardown_class(cls):
        pass
        # teardown_class() after any methods in this class

    def test_trading_type(self):
        from ..enums import TradingType
        t0 = TradingType.NONE
        t1 = TradingType.SANDBOX
        t2 = TradingType.LIVE
        t3 = TradingType.BACKTEST
        assert t0 == TradingType.NONE
        assert t1 == TradingType.SANDBOX
        assert t2 == TradingType.LIVE
        assert t3 == TradingType.BACKTEST
        assert t0 == TradingType('NONE')
        assert t1 == TradingType('SANDBOX')
        assert t2 == TradingType('LIVE')
        assert t3 == TradingType('BACKTEST')

    def test_exchange_type(self):
        from ..enums import ExchangeType, ExchangeType_from_string
        t0 = ExchangeType.NONE
        t1 = ExchangeType.COINBASE
        t2 = ExchangeType.GEMINI
        t3 = ExchangeType.KRAKEN
        t4 = ExchangeType.POLONIEX
        assert t0 == ExchangeType.NONE
        assert t1 == ExchangeType.COINBASE
        assert t2 == ExchangeType.GEMINI
        assert t3 == ExchangeType.KRAKEN
        assert t4 == ExchangeType.POLONIEX
        assert t0 == ExchangeType_from_string('NONE')
        assert t1 == ExchangeType_from_string('COINBASE')
        assert t2 == ExchangeType_from_string('GEMINI')
        assert t3 == ExchangeType_from_string('KRAKEN')
        assert t4 == ExchangeType_from_string('POLONIEX')

    def test_tick_type(self):
        from ..enums import TickType, TickType_from_string
        try:
            t0 = TickType_from_string('test')
            assert False
        except ValueError:
            pass

    def test_currency_type(self):
        from ..enums import CurrencyType
        t0 = CurrencyType.USD
        t1 = CurrencyType.BTC
        t2 = CurrencyType.ETH
        t3 = CurrencyType.LTC
        assert t0 == CurrencyType.USD
        assert t1 == CurrencyType.BTC
        assert t2 == CurrencyType.ETH
        assert t3 == CurrencyType.LTC
        assert t0 == CurrencyType('USD')
        assert t1 == CurrencyType('BTC')
        assert t2 == CurrencyType('ETH')
        assert t3 == CurrencyType('LTC')

    def test_side(self):
        from ..enums import Side
        t0 = Side.NONE
        t1 = Side.BUY
        t2 = Side.SELL
        assert t0 == Side.NONE
        assert t1 == Side.BUY
        assert t2 == Side.SELL
        assert t0 == Side('NONE')
        assert t1 == Side('BUY')
        assert t2 == Side('SELL')

    def test_order_type(self):
        from ..enums import OrderType
        t0 = OrderType.NONE
        t1 = OrderType.MARKET
        t2 = OrderType.LIMIT
        assert t0 == OrderType.NONE
        assert t1 == OrderType.MARKET
        assert t2 == OrderType.LIMIT
        assert t0 == OrderType('NONE')
        assert t1 == OrderType('MARKET')
        assert t2 == OrderType('LIMIT')

    def test_order_sub_type(self):
        from ..enums import OrderSubType
        t0 = OrderSubType.NONE
        t1 = OrderSubType.POST_ONLY
        t2 = OrderSubType.FILL_OR_KILL
        # t3 = OrderSubType.ALL_OR_NOTHING
        assert t0 == OrderSubType.NONE
        assert t1 == OrderSubType.POST_ONLY
        assert t2 == OrderSubType.FILL_OR_KILL
        # assert t3 == OrderSubType.ALL_OR_NOTHING
        assert t0 == OrderSubType('NONE')
        assert t1 == OrderSubType('POST_ONLY')
        assert t2 == OrderSubType('FILL_OR_KILL')
        # assert t3 == OrderSubType(3)



import json
from functools import lru_cache
from ..config import ExchangeConfig
from ..enums import ExchangeType
from ..exchange import Exchange
from ..structs import MarketData


class KrakenExchange(Exchange):
    def __init__(self, exchange_type: ExchangeType, options: ExchangeConfig) -> None:
        super(KrakenExchange, self).__init__(exchange_type, options)
        self._last = None
        self._orders = {}

    @lru_cache(None)
    def subscription(self):
        return [json.dumps({
            "event": "subscribe",
            "pair": [
                [self.currencyPairToString(x) for x in self.options().currency_pairs]
            ],
            "subscription": {
                "name": "ticker"
            }
        })]

    @lru_cache(None)
    def heartbeat(self):
        return ''

    def tickToData(self, jsn: dict) -> MarketData:
        raise NotImplementedError()



from typing import List
from .config import ExecutionConfig
from .enums import Side, TradingType, TradeResult
from .exchange import Exchange
from .logging import log
from .structs import TradeRequest, TradeResponse, Account


class Execution(object):
    def __init__(self, options: ExecutionConfig, exchanges: List[Exchange], accounts: List[Account]) -> None:
        self.trading_type = options.trading_type
        self.exchanges = exchanges
        self.accounts = accounts
        self._backtest_id = 1

    def insufficientFunds(self, req):
        resp = TradeResponse(request=req,
                             side=req.side,
                             exchange=req.exchange,
                             volume=0.0,
                             price=0.0,
                             instrument=req.instrument,
                             status=TradeResult.REJECTED,
                             time=req.time,
                             strategy=req.strategy,
                             order_id='')
        return resp

    def backtest(self, req):
        resp = TradeResponse(request=req,
                             side=req.side,
                             exchange=req.exchange,
                             volume=req.volume,
                             price=req.price,
                             instrument=req.instrument,
                             status=TradeResult.FILLED,
                             time=req.time,
                             strategy=req.strategy,
                             order_id=str(self._backtest_id))
        self._backtest_id += 1
        return resp

    def simulation(self, req):
        resp = TradeResponse(request=req,
                             side=req.side,
                             exchange=req.exchange,
                             volume=req.volume,
                             price=req.price,
                             instrument=req.instrument,
                             status=TradeResult.FILLED,
                             time=req.time,
                             strategy=req.strategy,
                             order_id=str(self._backtest_id))
        self._backtest_id += 1
        return resp

    def requestBuy(self, req: TradeRequest) -> TradeResponse:
        # can afford?
        balance = self.exchanges[req.exchange].accounts()[req.instrument.underlying.value[1]].balance

        if balance < req.volume * req.price:
            return self.insufficientFunds(req)

        if self.trading_type == TradingType.BACKTEST:
            return self.backtest(req)
        if self.trading_type == TradingType.SIMULATION:
            return self.simulation(req)

        resp = self.exchanges[req.exchange].buy(req)
        log.info('Order info: %s' % resp)
        return resp

    def requestSell(self, req: TradeRequest) -> TradeResponse:
        # can afford?
        balance = self.exchanges[req.exchange].accounts()[req.instrument.underlying.value[0]].balance

        if balance < req.volume:
            return self.insufficientFunds(req)

        if self.trading_type == TradingType.BACKTEST:
            return self.backtest(req)
        if self.trading_type == TradingType.SIMULATION:
            return self.simulation(req)

        resp = self.exchanges[req.exchange].sell(req)
        log.info('Order info: %s' % resp)
        return resp

    def request(self, req: TradeRequest) -> TradeResponse:
        if req.side == Side.BUY:
            return self.requestBuy(req)
        return self.requestSell(req)

    def cancel(self, resp: TradeResponse):  # TODO
        return self.exchanges[resp.exchange].cancel(resp)

    def cancelAll(self):
        for ex in self.exchanges.values():
            ex.cancelAll()



from datetime import datetime


class TestStructs:
    def setup(self):
        pass
        # setup() before each test method

    def teardown(self):
        pass
        # teardown() after each test method

    @classmethod
    def setup_class(cls):
        pass
        # setup_class() before any methods in this class

    @classmethod
    def teardown_class(cls):
        pass
        # teardown_class() after any methods in this class

    def test_to_dict(self):
        from ..structs import Instrument
        from ..enums import PairType, InstrumentType
        i = Instrument(underlying=PairType.BTCUSD)
        x = i.to_dict()
        print(x)
        assert x == {'type': InstrumentType.PAIR, 'underlying': PairType.BTCUSD}

        x = i.to_dict(True)
        assert x == {'type': 'InstrumentType.PAIR', 'underlying': 'BTC/USD'}

        x = i.to_dict(True, True)
        assert x == {'type': 'InstrumentType.PAIR', 'underlying': 'BTC/USD'}

    def test_MarketData(self):
        from ..structs import MarketData, Instrument
        from ..enums import TickType, Side, PairType, ExchangeType
        m = MarketData(time=datetime.now(),
                       volume=1.0,
                       price=1.0,
                       instrument=Instrument(underlying=PairType.BTCUSD),
                       type=TickType.TRADE,
                       exchange=ExchangeType.COINBASE,
                       side=Side.BUY)
        # TODO no fields yet
        assert m

    def test_TradeRequest(self):
        from ..structs import TradeRequest, Instrument
        from ..enums import Side, OrderType, PairType, ExchangeType
        t = TradeRequest(side=Side.BUY,
                         instrument=Instrument(underlying=PairType.BTCUSD),
                         order_type=OrderType.MARKET,
                         volume=1.0,
                         exchange=ExchangeType.COINBASE,
                         price=1.0,
                         time=datetime.now())
        assert t
        # side = Side
        # volume = float
        # price = float
        # exchange = ExchangeType
        # currency = CurrencyType
        # order_type = OrderType
        # order_sub_type = OrderSubType

    def test_TradeResponse(self):
        from ..structs import TradeRequest, TradeResponse, TradeResult, Instrument
        from ..enums import Side, OrderType, PairType, ExchangeType
        req = TradeRequest(side=Side.BUY,
                           order_type=OrderType.MARKET,
                           exchange=ExchangeType.COINBASE,
                           instrument=Instrument(underlying=PairType.BTCUSD),
                           volume=1.0,
                           price=1.0,
                           time=datetime.now())
        t = TradeResponse(request=req,
                          side=Side.BUY,
                          instrument=Instrument(underlying=PairType.BTCUSD),
                          volume=0.0,
                          price=0.0,
                          status=TradeResult.FILLED,
                          exchange=ExchangeType.COINBASE,
                          time=datetime.now(),
                          order_id='1')
        assert t
        # side = Side
        # volume = float
        # price = float
        # success = bool



from functools import lru_cache
from .enums import ExchangeType, TradingType


EXCHANGE_MARKET_DATA_ENDPOINT = lru_cache(None)(lambda name, typ: {  # noqa: E731
    (ExchangeType.COINBASE, TradingType.SANDBOX): 'wss://ws-feed-public.sandbox.pro.coinbase.com',
    (ExchangeType.COINBASE, TradingType.LIVE): 'wss://ws-feed.pro.coinbase.com',
    (ExchangeType.COINBASE, TradingType.SIMULATION): 'wss://ws-feed.pro.coinbase.com',

    (ExchangeType.GEMINI, TradingType.SANDBOX): 'wss://api.sandbox.gemini.com/v1/marketdata/%s?heartbeat=true',
    (ExchangeType.GEMINI, TradingType.LIVE): 'wss://api.gemini.com/v1/marketdata/%s?heartbeat=true',
    (ExchangeType.GEMINI, TradingType.SIMULATION): 'wss://api.gemini.com/v1/marketdata/%s?heartbeat=true',

    (ExchangeType.KRAKEN, TradingType.SANDBOX): 'wss://ws-beta.kraken.com',
    (ExchangeType.KRAKEN, TradingType.LIVE): 'wss://ws.kraken.com',
    (ExchangeType.KRAKEN, TradingType.SIMULATION): 'wss://ws.kraken.com',

    (ExchangeType.POLONIEX, TradingType.SANDBOX): '',
    (ExchangeType.POLONIEX, TradingType.LIVE): 'wss://api2.poloniex.com',
    (ExchangeType.POLONIEX, TradingType.SIMULATION): 'wss://api2.poloniex.com',

    # (ExchangeType.DERIBIT, TradingType.SANDBOX): 'wss://test.deribit.com/ws/api/v1/',
    # (ExchangeType.DERIBIT, TradingType.LIVE): 'wss://www.deribit.com/ws/api/v1/',
    # (ExchangeType.DERIBIT, TradingType.SIMULATION): 'wss://www.deribit.com/ws/api/v1/',

}.get((name, typ), None))



from ..strategy import TradingStrategy
from ..structs import MarketData, TradeRequest, TradeResponse
from ..enums import Side, OrderType
from ..logging import log


class BuyAndHoldStrategy(TradingStrategy):
    def __init__(self, *args, **kwargs) -> None:
        super(BuyAndHoldStrategy, self).__init__(*args, **kwargs)
        self.bought = {}

    def onFill(self, res: TradeResponse) -> None:
        self.bought[res.instrument] = res
        log.info('bought %.2f @ %.2f' % (res.volume, res.price))

    def onTrade(self, data: MarketData) -> bool:
        if data.instrument not in self.bought:
            req = TradeRequest(side=Side.BUY,
                               volume=1,
                               instrument=data.instrument,
                               order_type=OrderType.MARKET,
                               exchange=data.exchange,
                               price=data.price,
                               time=data.time)
            log.info("requesting buy : %s", req)
            self.request(req)
            self.bought[data.instrument] = 'pending'

    def onError(self, e) -> None:
        log.critical(e)

    def onChange(self, data: MarketData) -> None:
        pass

    def onCancel(self, data: MarketData) -> None:
        pass

    def onOpen(self, data: MarketData) -> None:
        pass

    def slippage(self, resp: TradeResponse) -> TradeResponse:
        slippage = resp.price * .0001  # .01% price impact
        if resp.side == Side.BUY:
            # price moves against (up)
            resp.slippage = slippage
            resp.price += slippage
        else:
            # price moves against (down)
            resp.slippage = -slippage
            resp.price -= slippage
        return resp

    def transactionCost(self, resp: TradeResponse) -> TradeResponse:
        txncost = resp.price * resp.volume * .0025  # gdax is 0.0025 max fee
        if resp.side == Side.BUY:
            # price moves against (up)
            resp.transaction_cost = txncost
            resp.price += txncost
        else:
            # price moves against (down)
            resp.transaction_cost = -txncost
            resp.price -= txncost
        return resp


class BuyAndHoldStrategy2(TradingStrategy):
    def __init__(self, *args, **kwargs) -> None:
        super(BuyAndHoldStrategy2, self).__init__(*args, **kwargs)
        self.sold = None

    def onFill(self, res: TradeResponse) -> None:
        self.sold = res
        log.info('d->g:sold %.2f @ %.2f' % (res.volume, res.price))

    def onTrade(self, data: MarketData) -> bool:
        if self.sold is None:
            req = TradeRequest(side=Side.SELL,
                               volume=.1,
                               instrument=data.instrument,
                               order_type=OrderType.LIMIT,
                               exchange=data.exchange,
                               price=data.price,
                               time=data.time)
            log.info("requesting sell : %s", req)
            self.request(req)
            self.sold = 'pending'

    def onError(self, e) -> None:
        log.critical(e)

    def onChange(self, data: MarketData) -> None:
        pass

    def onCancel(self, data: MarketData) -> None:
        pass

    def onOpen(self, data: MarketData) -> None:
        pass



import pandas as pd
from .config import BacktestConfig
from .data_source import StreamingDataSource
from .logging import log
from .structs import MarketData, Instrument
from .enums import PairType, TickType, ExchangeType_from_string, Side


def line_to_data(record):
    data = MarketData(time=record.name[0],
                      volume=record.volume,
                      price=record.close,
                      type=TickType.TRADE,
                      instrument=Instrument(underlying=PairType.from_string(record.name[1])),
                      exchange=ExchangeType_from_string(record.exchange),
                      side=Side.NONE)
    return data


class Backtest(StreamingDataSource):
    def __init__(self, options: BacktestConfig) -> None:
        super(Backtest, self).__init__()

    def run(self, engine) -> None:
        log.info('Starting....')

        datas = [ex.historical() for ex in engine.exchanges.values()]
        data = pd.concat(datas)
        data.sort_index()

        for index, row in data.iterrows():
            self.receive(line_to_data(row))
        log.info('Backtest done, running analysis.')

        self.callback(TickType.ANALYZE, engine)
        log.info('Analysis completed.')

    def receive(self, data: MarketData) -> None:
        # TODO allow if market data for bid/ask
        if data.type == TickType.TRADE:
            self.callback(TickType.TRADE, data)
        else:
            self.callback(TickType.ERROR, data)

    def close(self) -> None:
        pass

    def seqnum(self, num: int) -> None:
        pass

    def tickToData(self, tick: str) -> None:
        pass



from datetime import datetime
from typing import List
from .config import RiskConfig
from .enums import Side, TradeResult, OrderType, RiskReason, ExchangeType
from .exchange import Exchange
from .logging import log
from .structs import TradeRequest, TradeResponse, Account, Instrument
from .utils import iterate_accounts


class Risk(object):
    def __init__(self, options: RiskConfig, exchanges: List[Exchange], accounts: List[Account]) -> None:
        self.trading_type = options.trading_type
        self.max_drawdown = options.max_drawdown
        self.max_risk = options.max_risk
        self.total_funds = options.total_funds

        self.outstanding = 0.0
        self.starting_funds = options.total_funds
        self.drawdown = 0.0

        self.exchanges = exchanges
        self.accounts = accounts

    def _constructResp(self,
                       side: Side,
                       exchange: ExchangeType,
                       instrument: Instrument,
                       order_type: OrderType,
                       vol: float,
                       price: float,
                       time: datetime,
                       status: bool,
                       strat: object,
                       reason: RiskReason) -> TradeRequest:
        resp = TradeRequest(side=side,
                            exchange=exchange,
                            instrument=instrument,
                            order_type=order_type,
                            volume=vol,
                            price=price,
                            time=time,
                            risk_check=status,
                            risk_reason=reason,
                            strategy=strat,
                            )
        return resp

    def request(self, req: TradeRequest) -> TradeRequest:
        # update risk as soon as order will go out
        log.info('Requesting %f @ %f', req.volume, req.price)
        total = req.volume * req.price
        total = total * -1 if req.side == Side.SELL else total
        max = self.max_risk / 100.0 * self.total_funds

        if (total + self.outstanding) <= max:
            # room for full volume
            log.info('Risk check passed for order: %s' % req)
            return self._constructResp(side=req.side,
                                       exchange=req.exchange,
                                       instrument=req.instrument,
                                       order_type=req.order_type,
                                       vol=req.volume,
                                       price=req.price,
                                       time=req.time,
                                       status=True,
                                       strat=req.strategy,
                                       reason=RiskReason.NONE)

        elif self.outstanding < max:
            # room for some volume
            volume = (max - self.outstanding) / req.price
            log.info('Risk check passed for partial order: %s' % req)
            return self._constructResp(side=req.side,
                                       exchange=req.exchange,
                                       instrument=req.instrument,
                                       order_type=req.order_type,
                                       vol=volume,
                                       price=req.price,
                                       time=req.time,
                                       status=True,
                                       strat=req.strategy,
                                       reason=RiskReason.PARTIAL)

        # no room for volume
        log.info('Risk check failed for order: %s' % req)
        return self._constructResp(side=req.side,
                                   exchange=req.exchange,
                                   instrument=req.instrument,
                                   order_type=req.order_type,
                                   vol=0.0,
                                   price=req.price,
                                   time=req.time,
                                   status=False,
                                   strat=req.strategy,
                                   reason=RiskReason.FULL)

    def requestBuy(self, req: TradeRequest):
        '''precheck for risk compliance'''
        return self.request(req)

    def requestSell(self, req: TradeRequest):
        '''precheck for risk compliance'''
        return self.request(req)

    def updateAccounts(self):
        '''update risk numbers'''
        log.critical('risk not fully implemented - updateRisk')
        value = 0.0

        for account in iterate_accounts(self.accounts):
            value += account.value

        if value < self.total_funds:
            log.info(f'restricting total funds from {self.total_funds} to {value}')
        elif value > self.total_funds:
            log.info(f'expanding total funds from {self.total_funds} to {value}')

        self.total_funds = value
        log.info(f'drawdown: {self.total_funds - self.starting_funds}')

    def update(self, resp: TradeResponse):
        '''update risk after execution'''
        log.critical('risk not fully implemented - update')
        self.updateAccounts()

        if resp.status == TradeResult.FILLED:
            # FIXME
            self.outstanding += abs(resp.volume * resp.price) * (1 if resp.side == Side.BUY else -1)

        elif resp.status == TradeResult.REJECTED:
            self.outstanding -= abs(resp.volume * resp.price) * (1 if resp.side == Side.BUY else -1)

    def cancel(self, resp: TradeResponse):
        '''update risk after cancelling or rejecting order'''
        log.critical('risk not fully implemented - cancel')
        self.updateAccounts()

        self.outstanding -= abs(resp.volume * resp.price) * (1 if resp.side == Side.BUY else -1)



import ccxt
import pandas as pd
import time
from datetime import datetime
from functools import lru_cache
from numpy.random import normal
from random import randint, random, choice  # noqa: F401
from typing import List
from ..config import SyntheticExchangeConfig
from ..enums import ExchangeType, TickType, Side, PairType, OrderType, CurrencyType, TradingType, ExchangeType_to_string
from ..exchange import Exchange
from ..order_book import OrderBook
from ..structs import MarketData, TradeRequest, TradeResponse, Instrument, Account


class SyntheticExchange(Exchange):
    def __init__(self,
                 exchange_type: ExchangeType,
                 options: SyntheticExchangeConfig,
                 query_engine=None) -> None:
        super(SyntheticExchange, self).__init__(exchange_type=exchange_type, options=options, query_engine=query_engine)
        self._book = OrderBook(options.instruments)
        self._instruments = options.instruments
        self._trading_type = options.trading_type

        # random walk generated by laplace
        # generated order book will skew in this direction
        self._direction = options.direction

        # generated order book will skew with this volatility
        self._volatility = options.volatility

        self._starting_price = 100

        self._underlying_exchange = options.exchange_type
        self._advesaries = []

    def generateMsg(self):
        price = 1000
        while True:
            # take a random walk

            # endpoint
            next_ = normal(self._direction, self._volatility)

            # walk between
            # substeps = normal(self._direction/10, self._volatility/10, randint(1, 10))

            # scale to fill next
            # substeps = substeps*(next_/substeps.cumsum()[-1])
            # for step in substeps:
            #     price += step
            #     price = abs(price)
            price = abs(price + next_)
            data = MarketData(time=datetime.now(),
                              volume=random() * 10,
                              price=price,
                              type=TickType.TRADE,
                              instrument=self._instruments[0],
                              side=choice([Side.BUY, Side.SELL]),
                              exchange=self._underlying_exchange,
                              remaining=random() * 10,
                              sequence=-1,
                              order_type=OrderType.NONE)
            if self._trading_type != TradingType.BACKTEST:
                time.sleep(random() * 3)
            else:
                time.sleep(random() / 500)
            yield data

    async def run(self, trading) -> None:
        while True:
            await self.receive()

    async def receive(self) -> None:
        for msg in self.generateMsg():
            self.callback_data(msg)

    @lru_cache(None)
    def oe_client(self):
        return ccxt.coinbasepro({'enableRateLimit': True})

    async def close(self) -> None:
        '''close the websocket'''
        pass

    @lru_cache(None)
    def accounts(self):
        all_curs = set()
        for inst in self._options.instruments:
            all_curs.add(inst.underlying.value[0])
            all_curs.add(inst.underlying.value[1])

        return [Account(id=str(cur.value),
                        currency=cur,
                        balance=100,
                        exchange=self.exchange(),
                        value=-1,
                        asOf=datetime.now()) for cur in all_curs]

    @lru_cache(None)
    def currencies(self) -> List[CurrencyType]:
        return [CurrencyType(x) for x in self.oe_client().fetch_curencies()]

    @lru_cache(None)
    def markets(self) -> List[Instrument]:
        return [Instrument(underlying=PairType.from_string(m['symbol'])) for m in self.oe_client().fetch_markets()]

    def orderBook(self, level=1):
        '''get order book'''
        raise NotImplementedError()

    def buy(self, req: TradeRequest) -> TradeResponse:
        '''execute a buy order'''
        for ad in self._advesaries:
            req = ad.beforeOrder(req)

        raise NotImplementedError()

        for ad in self._advesaries:
            req = ad.afterOrder(req)

    def sell(self, req: TradeRequest) -> TradeResponse:
        '''execute a sell order'''
        for ad in self._advesaries:
            req = ad.beforeOrder(req)

        raise NotImplementedError()

        for ad in self._advesaries:
            req = ad.afterOrder(req)

    def cancel(self, resp: TradeResponse):
        for ad in self._advesaries:
            resp = ad.beforeOrder(resp)

        raise NotImplementedError()

        for ad in self._advesaries:
            resp = ad.afterOrder(resp)

    def cancelAll(self, resp: TradeResponse):
        pass

    @lru_cache(None)
    def subscription(self):
        return []

    @lru_cache(None)
    def heartbeat(self):
        return []

    def tickToData(self, jsn: dict) -> MarketData:
        pass

    def historical(self):
        data = []
        i = 0
        for msg in self.generateMsg():
            # msgs = asyncio.get_event_loop().run_until_complete(asyncio.ensure_future(self.generateMsg()))
            data.append(msg)

            i += 1
            if i > 1000:
                break

        data = [{'timestamp': d.time,
                 'open': d.price,
                 'high': d.price,
                 'low': d.price,
                 'close': d.price,
                 'volume': d.volume} for d in data]

        dfs = [{'pair': str(self._instruments[0].currency_pair),
                'exchange': ExchangeType_to_string(self._underlying_exchange),
                'data': data}]
        df = pd.io.json.json_normalize(dfs, 'data', ['pair', 'exchange'])
        df['timestamp'] = pd.to_datetime(df['timestamp'], unit='ms')
        df.set_index(['timestamp', 'pair'], inplace=True)
        df.sort_index(inplace=True)
        return df


class Adversary(object):
    def __init__(self):
        pass

    def beforeData(self, data: MarketData, orig: MarketData) -> MarketData:
        pass

    def afterData(self, market) -> None:
        pass

    def beforeOrder(self, data: TradeRequest, market) -> None:
        pass

    def afterOrder(self, data: TradeResponse, market) -> None:
        pass



import ccxt
import logging
import os
import pytz
from datetime import datetime
from functools import lru_cache
from .enums import ExchangeType, ExchangeType_from_string, ExchangeTypes, CurrencyType, OrderType, Side, PairType
from .exceptions import AATException
from .logging import log


@lru_cache(100)
def parse_date(indate: str) -> datetime:
    '''parse date

    Args:
        indate (string, int, or datetime): input to convert to datetime
    Returns:
        datetime
    '''
    if isinstance(indate, datetime):
        return indate
    try:
        date = datetime.utcfromtimestamp(float(indate))
        date = pytz.utc.localize(date).astimezone(pytz.timezone('EST')).replace(tzinfo=None)
    except ValueError:
        date = datetime.strptime(indate, "%Y-%m-%dT%H:%M:%S.%fZ")
    return date


@lru_cache(None)
def ex_type_to_ex(ex: ExchangeType):
    '''Convert Exchange type to Exchange class

    Args:
        ex (ExchangeType): exchange type
    Returns:
        Exchange
    '''
    if ex == ExchangeType.COINBASE:
        from .exchanges.coinbase import CoinbaseExchange
        return CoinbaseExchange
    elif ex == ExchangeType.GEMINI:
        from .exchanges.gemini import GeminiExchange
        return GeminiExchange
    elif ex == ExchangeType.KRAKEN:
        from .exchanges.kraken import KrakenExchange
        return KrakenExchange
    elif ex == ExchangeType.SYNTHETIC:
        from .exchanges.synthetic import SyntheticExchange
        return SyntheticExchange
    raise NotImplementedError(f'Exchange type not implemented : {ex} ')


@lru_cache(None)
def get_keys_from_environment(prefix: str) -> tuple:
    '''get exchange keys from environment variables

    Args:
        prefix (str): exchange name e.g. COINBASE
    Returns
        key (str): private key
        secret (str): private secret
        passphrase (str): optional passphrase
    '''

    prefix = prefix.upper()
    key = os.environ[prefix + '_API_KEY']
    secret = os.environ[prefix + '_API_SECRET']
    passphrase = os.environ[prefix + '_API_PASS']
    return key, secret, passphrase


@lru_cache(None)
def str_to_side(s: str) -> Side:
    s = s.upper()
    if 'BUY' in s or 'BID' in s:
        return Side.BUY
    if 'SELL' in s or 'ASK' in s:
        return Side.SELL
    return Side.NONE


@lru_cache(None)
def str_to_order_type(s: str) -> OrderType:
    s = s.upper()
    if 'MARKET' in s:
        return OrderType.MARKET
    if 'LIMIT' in s:
        return OrderType.LIMIT
    return OrderType.NONE


@lru_cache(None)
def str_to_exchange(exchange: str) -> ExchangeType:
    if exchange.upper() not in ExchangeTypes:
        raise AATException(f'Exchange not recognized: {exchange}')
    return ExchangeType_from_string(exchange.upper())


def set_verbose(level):
    # Print/log extra info
    if level >= 2:
        log.setLevel(logging.DEBUG)
    elif level == 1:
        log.setLevel(logging.INFO)
    log.info('running in verbose mode!')


def trade_req_to_params(req) -> dict:
    ret = {}
    ret['symbol'] = str(req.instrument.underlying)
    ret['side'] = 'buy' if req.side == Side.BUY else 'sell'
    ret['type'] = 'market' if req.order_type == OrderType.MARKET else 'limit'
    ret['amount'] = req.volume

    if ret['type'] == 'limit':
        ret['price'] = req.price
    return ret


def exchange_type_to_ccxt_client(exchange_type):
    if exchange_type == ExchangeType.COINBASE:
        return ccxt.coinbasepro
    elif exchange_type == ExchangeType.GEMINI:
        return ccxt.gemini
    elif exchange_type == ExchangeType.KRAKEN:
        return ccxt.kraken
    elif exchange_type == ExchangeType.POLONIEX:
        return ccxt.poloniex


def tradereq_to_ccxt_order(req) -> dict:
    # TODO order_sub_type
    return dict(
        symbol=str(req.instrument.underlying),
        type=req.order_type.value.lower(),
        side=req.side.value.lower(),
        price=req.price,
        amount=req.volume,
        params={})


def findpath(inst, markets):
    '''find a path from left side of instrument to right side of instrument
    given the available markets

    Args:
        inst: Instrument with an underlying pairtype
        markets: List of Instruments
    Returns:
        tuple:
            instrument1, instrument2, invert pair1, invert pair2
    '''
    from .structs import Instrument

    # should do dijkstras to get cheapest path but im lazy
    to_try = (CurrencyType.BTC, CurrencyType.ETH)

    for base in to_try:
        currency_from = inst.underlying.value[0]
        currency_to = inst.underlying.value[1]
        c1_btc = Instrument(underlying=PairType((currency_from, base)))
        c1_btc_inv = Instrument(underlying=PairType((base, currency_from)))
        c2_btc = Instrument(underlying=PairType((base, currency_to)))
        c2_btc_inv = Instrument(underlying=PairType((currency_to, base)))

        if currency_from == base or \
           currency_to == base or \
           (c1_btc not in markets and c1_btc_inv not in markets) or \
           (c2_btc not in markets and c2_btc_inv not in markets):
            continue

        return {(True, True): (c1_btc_inv, c2_btc_inv, True, True),
                (True, False): (c1_btc_inv, c2_btc, True, False),
                (False, True): (c1_btc, c2_btc_inv, False, True),
                (False, False): (c1_btc, c2_btc, False, False)}.get((c1_btc not in markets,
                                                                     c2_btc not in markets))
    raise AATException(f'Need {"/".join(c.value for c in to_try)} for intermediary: {inst}')


def iterate_accounts(accounts_dict):
    for currency in accounts_dict:
        if not isinstance(currency, CurrencyType):
            continue
        for account in accounts_dict[currency].values():
            yield account


def sign(x): return (1, -1)[x < 0]


class pnl_helper(object):
    def __init__(self):
        self._records = []
        self._pnl = 0.0
        self._px = None
        self._volume = None
        self._realized = 0.0
        self._avg_price = self._px

    def price(self, px):
        self._pnl = (self._volume * (px - self._avg_price))
        self._records.append({'volume': self._volume, 'px': px, 'apx': self._avg_price, 'pnl': self._pnl + self._realized, 'unrealized': self._pnl, 'realized': self._realized})

    def exec(self, amt, px, side):
        if self._px is None:
            self._px = px
            self._volume = amt
            self._avg_price = self._px
            self._records.append({'volume': self._volume, 'px': px, 'apx': self._avg_price, 'pnl': self._pnl + self._realized, 'unrealized': self._pnl, 'realized': self._realized})
            return
        amt = amt if side == Side.BUY else -amt

        if sign(amt) == sign(self._volume):
            # increasing position
            self._avg_price = (self._avg_price * self._volume + px * amt) / (self._volume + amt)
            self._volume += amt
            self._pnl = (self._volume * (px - self._avg_price))
        else:
            if abs(amt) > abs(self._volume):
                # do both
                diff = self._volume
                self._volume = amt + self._volume

                # take profit/loss
                self._realized += (diff * (px - self._avg_price))

                # increasing position
                self._avg_price = px

            else:
                # take profit/loss
                self._volume += amt

                self._pnl = (self._volume * (px - self._avg_price))
                self._realized += (amt * (self._avg_price - px))
        self._records.append({'volume': self._volume, 'px': px, 'apx': self._avg_price, 'pnl': self._pnl + self._realized, 'unrealized': self._pnl, 'realized': self._realized})
